/*--------------------------------------------------------------------------

  SPARCAUDIO.C

  Play signal through sparc audio device

  Andrew Simpson, 22 January 1994

  Conversion to ulaw routines taken from sfsdac.c by Mark Huckvale.

---------------------------------------------------------------------------*/

#define PROGRAM "audio.c"

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioccom.h>
#include <sun/audioio.h>
#include <malloc.h>
#include <memory.h>
#include <math.h>
#define TRUE 1
#define FALSE 0

/* same values as myge.c */
#define SPEAKER 1
#define PHONES 2
#define LINE 4

/*-------------------------------------------------------------------------*/

/* Function Prototypes */

int openAudioDevice();
int outputAudioData(short *data, int noSamples, int sampleRate,
		     double gain, unsigned port);
void closeAudioDevice();
int nearestSunSampleRate (int srate);
unsigned char pcm2ulaw (short in);

/*-------------------------------------------------------------------------*/

/* Global Variables */

static int audiofd=0,hardwareType;
static audio_info_t audioInfo;

/*-------------------------------------------------------------------------*/

int openAudioDevice()

{
  /* Don't bother if already open */
  if (audiofd!=0) return(0);

  /* Open audio device */
  if ((audiofd = open("/dev/audio", O_WRONLY | O_NDELAY))<0)
    {
      if (errno == EBUSY)
	fprintf(stderr,"audio device is busy\n");
      else
	fprintf(stderr,"couldn't open the audio device\n");
      return(0);
    }

  /* Establish the type of hardware being used.

     A Sparc10 has a DBRI interface (man dbri), and can playback at
     a number (albeit limited) of different sample rates, and with
     16-bit precision.
	   
     Sparc1, Sparc2, and IPC Computers have an AMD chip which only
     allows playback at 8kHz, and with pseudo-12 bit encoding, 
     achieved using u-law encoding which squeezes each 12bit sample
     into 8bits.

     */

  /* get hardware type */
  ioctl (audiofd,AUDIO_GETDEV,&hardwareType);

  return(audiofd);
}
	
/*----------------------------------------------------------------------*/

int outputAudioData(short *data, int noSamples, int sampleRate,
		     double gain, unsigned port)

{
  int outputSamples;
  int actualSampleRate=0;
  unsigned sunPort;
  unsigned char *ulawPtr;
  long	stepSize, offset;
  static unsigned char *ulawData;

  if (audiofd==0)
    if (openAudioDevice()==0) 
      return(0);

  gain*=AUDIO_MAX_GAIN;
  sunPort=0;
  if (port & SPEAKER) sunPort |= AUDIO_SPEAKER;
  if (port & LINE) sunPort |= AUDIO_LINE_OUT;
  if (port & PHONES) sunPort |= AUDIO_HEADPHONE;

  /* get current settings */
  ioctl (audiofd,AUDIO_GETINFO,&audioInfo);

  if (hardwareType==AUDIO_DEV_SPEAKERBOX)
    {
      actualSampleRate=nearestSunSampleRate(sampleRate);

      /* 16bit linear encoding */
      audioInfo.play.precision = 16;
      audioInfo.play.channels = 1;
      audioInfo.play.sample_rate = actualSampleRate;
      audioInfo.play.encoding = AUDIO_ENCODING_LINEAR;
      audioInfo.play.gain=gain;
      audioInfo.play.port=sunPort;
      ioctl (audiofd,AUDIO_SETINFO,&audioInfo);

      write (audiofd,(char *)data,noSamples*sizeof(short));
      ioctl(audiofd,AUDIO_DRAIN,audiofd);
    }
  else
    {
      /* 8bit ulaw encoding, 8Khz */
      actualSampleRate=sampleRate;
      audioInfo.play.sample_rate = 8000;
      audioInfo.play.precision = 8;
      audioInfo.play.channels = 1;
      audioInfo.play.gain=gain;
      audioInfo.play.port=sunPort;
      audioInfo.play.encoding = AUDIO_ENCODING_ULAW;
      ioctl (audiofd,AUDIO_SETINFO,&audioInfo);

      /* allocate some memory for the conversion */
      if (ulawData==NULL)
	{
	  if ((ulawData=(unsigned char *)malloc(noSamples))==NULL)
	      return(0);
	}
      else 
	if ((ulawData=(unsigned char *)realloc((char *)ulawData,noSamples))==NULL)
	  return(0);

      /* calculate step size - 8 bit fixed point */
      stepSize = (long)(0.5+256.0*(double)sampleRate/8000.);
      
      /* now do conversion on the fly */
      for (offset=0, ulawPtr=ulawData, outputSamples=0;
	   (offset >> 8) < noSamples;
	   offset+=stepSize, ulawPtr++, outputSamples++)
	*ulawPtr = pcm2ulaw(data[offset >> 8]);

      /* and now play it out */
      write (audiofd,(char *)ulawData,outputSamples);
      ioctl(audiofd,AUDIO_DRAIN,audiofd);
    }
  return(actualSampleRate);
}

/*----------------------------------------------------------------------*/

void closeAudioDevice()
     
{
  if (audiofd!=0)
    {
      close(audiofd);
      audiofd=0;
    }
}

/*----------------------------------------------------------------------*/

/* thanks to Mark Huckvale for this routine */

/* Sun SPARC-10 dbri playback frequency table */
static int sunfrq[]=
{8000,9600,11025,16000,18900,22050,32000,37800,44100,48000,0};

/* get nearest sample rate */
int nearestSunSampleRate (int srate)

{
  int	i;
  int	min,idx;
  
  idx=0;
  min=abs(sunfrq[0]-srate);
  for (i=0;sunfrq[i];i++) {
    if (abs(sunfrq[i]-(int)srate)<min) {
      min=abs(sunfrq[i]-(int)srate);
      idx=i;
    }
  }
  return(sunfrq[idx]);
}


/*----------------------------------------------------------------------*/

/* Thanks to Mark Huckvale for this routine from sfsdac.c*/

/* U-law sample conversion from 12 to 8 bits */	
#define	UMAX      ((1 << 13) - 1)

unsigned char pcm2ulaw (short in)

{
  register unsigned char out;
  
  /* convert to 12 bit */
  in >>= 4;

  if (in < 0) {
    out = (unsigned char) 0;
    in = -in;
  }
  else
    out = (unsigned char) 0x80;
  
  in += 33;
  
  if (in > UMAX) in = UMAX;
  
  if (in < 0x200) {
    if (in < 0x080) {
      if (in < 0x040) {
	if (in < 0x20)
	  in = 0;
	else
	  in = (in >> 1) ^ 0x10;
      }
      else
	in = (in >> 2);
    }
    else {
      if (in < 0x100)
	in = (in >> 3) ^ 0x30;
      else
	in = (in >> 4) ^ 0x20;
    }
  }
  else {
    if (in < 0x800) {
      if (in < 0x400) 
	in = (in >> 5) ^ 0x50;
      else
	in = (in >> 6) ^ 0x40;
    }
    else {
      if (in < 0x1000)
	in = (in >> 7) ^ 0x70;
      else
	in = (in >> 8) ^ 0x60;
    }
  }
  out |= (unsigned char) in;
  out = ~out;  
  return (out);
}

/*----------------------------------------------------------------------*/
