/*----------------------------------------------------------------------------
  
  KPE - KLATT PARAMETER EDITOR
  
  Andrew Simpson
  Department of Linguistics and Phonetics
  University College London
  Wolfson House
  4 Stephenson Way
  London NW1 2HE
  
  andrew@phon.ucl.ac.uk
  
  August 3 1995
  
  Copyright University College London
  
 ----------------------------------------------------------------------------*/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include "suit.h"
#include <sys/types.h>
#include <sys/time.h>

#ifdef SFS
#include "sfs.h"
#endif

/*--------------------------------------------------------------------------*/

#ifdef KLATT88
#define PROGRAM "kpe88"
#else
#define PROGRAM "kpe80"
#endif

#define MAXIMUM(x,y) (((x)>(y))?(x):(y))
#define MAXSAMPLE 32767.0
#define LINEWIDTH 2

/* Memory allocation */
#define SHORT 1
#define DOUBLE 2
#define FLOAT 3
#define CHAR 4

#ifdef SUNOS413
/* Sound output */
#define SPEAKER 1
#define PHONES 2
#define LINE 4
#endif

/* Spectrum/Spectrogram calculation */
#define SPECRANGE 60.0
#define HEADROOM 15.0
#define NOSPECTRA 750
#define FFTSIZE 1024
#define WINDOWLENGTH 0.006

/* maximum number of points that can be drawn */
#define MAXPOINTS 10000

/* Parameter State Constants */
#define NOTDISPLAYED 0
#define DISPLAYED 1
#define BEINGEDITED 2

/* Various new object attributes */
#define STATE "state"
#define NUMBER "number"
#define EDITEDPARAM "editedParam"
#define DISPLAY "display"
#define GAIN "gain"
#define OUTPUT "output"
#define PARAMNAME "paramname"
#define PARAMDESC "paramdesc"
#define TIME "time"
#define PARAMETERVALUE "parameter_value"
#define CALLBACKARGUMENT "callback_parameter"
#define DISPLAYMODE "displayMode"
#define PARAMETERS 1
#define SPECTRA 2
#define NOFRAMES "noFrames"
#define START "start"
#define END "end"
#define DEFAULTWINDOW "defaultWindow"
#define WRITEMODE "writeMode"
#define SGRAM "sgram"

/* Object types */
#define STACKER 1
#define PULLDOWNMENU 2
#define BUTTON 3
#define MYLABEL 4
#define ONOFFSWITCH 5
#define BOUNDEDVALUE 6
#define ARROWBUTTON 7

/* file types */
#define SFSFILE 1
#define WAVFILE 2
#define AUFILE 3

/* new line */
#define NL 10

/*--------------------------------------------------------------------------*/

/* Information about the two waveforms */

static struct waveInfo {
  char *name;
  int defineWindow;
  GP_color colour;
  SUIT_object display;
  SUIT_window window;
  
  /* waveform variables */
  int used;
  short *waveS;
  double *waveD;
  int recalcSpec;
  
  /* Spectrum Variables */
  int specSize;
  double *spectrum;
  double *specX;
  
  /* Spectrogram Variables */
  int fftSize;
  float windowDuration;
  char *greyScales;
  float *sgram;
  double *sgramX,*sgramY;
  int recalcSgram;
  int recalcGreyScales;
  
  /* Gain Control */
  SUIT_object gainControl;
  SUIT_object startTime;
  SUIT_object endTime;
  SUIT_object fileName;
} wfi[] = {{"original",FALSE},{"synthesised",TRUE},{NULL}};

#define ORIG 0
#define SYNTH 1
#define BOTH 2

/*--------------------------------------------------------------------------*/

/* EXTERNAL FUNCTIONS */

/* audio.c */
extern void openAudioDevice();
extern void closeAudioDevice();
#ifdef SUNOS413
extern int nearestSunSampleRate (int srate);
#endif
extern int outputAudioData(short *data, int noSamples, int sampleRate,
			   double gain,int port);

/* sigproc.c */
extern void forwardFFT (float *data, int n);
extern void ffta (float *data, int n);
extern void ffts (float *data, int n);
extern void realft(float *data, int n, int isign);
extern void four1(float *data, int nn, int isign);
extern void fxcep (short *waveform, int sampleRate, int noFrames, 
		   float frameStep, float *fxContour);
/* resample.c */
extern void intdecE(float *x, int xlen, int fx,
		    float **y, int *ylen, int fy);

/*--------------------------------------------------------------------------*/

/* FUNCTION PROTOTYPES */

void createObjects ();
void initialise (int noFrames, int updateInterval, int sampleRate,
		 int reset, int synthesise, int display);
void myAllocate (void **pointerPtr, size_t nelem, int type);

void redrawCallback (SUIT_object object);
void variedParamButtonCallback (SUIT_object object);
void fixedParamButtonCallback (SUIT_object object);
void resample (int, int, double **);
void parameterPaintProc (SUIT_object object);
void parameterHitProc (SUIT_object display, SUIT_event myEvent);
void displayPosition (SUIT_object, GP_point);

void waveformPaintProc (SUIT_object object);
void zoomCallback (SUIT_object button);
void unZoomCallback (SUIT_object button);
void rescaleCallback (SUIT_object button);
void waveformHitProc (SUIT_object display, SUIT_event myEvent);
void spectrumCallback (SUIT_object object);
void calculateDFT (int smoothed);
float calculateSpectrogram (struct waveInfo *wpi);
void calculateGreyscales (struct waveInfo *wpi, float dBcorrecton);
void fxCallback (SUIT_object object);

void synthesiseCallback (SUIT_object object);
void new (SUIT_object object);
void writeRIFFwaveheader(FILE *op, int nsample, double srate, int nchan,int do8bit);
void writeRIFFheader(FILE *op, int nsample, int nchan, int do8bit);
void loadAudioFile (SUIT_object object);
void saveSfsFile (SUIT_object object, char *passedFilename);
short readShort (FILE *fp, int lsbFirst);
long readLong (FILE *fp, int lsbFirst);
int resampleWave (int sr, int newSr, short **wave, int noSamples);
void loadParFile (SUIT_object object, char *passedFilename);
void saveParFile (SUIT_object object);
void hideAllCallback (SUIT_object object);
void defaultValueCallback (SUIT_object object);
void memoriseParamCallback (SUIT_object object);
void recallParamCallback (SUIT_object object);
void memoriseAllCallback (SUIT_object object);
void recallAllCallback (SUIT_object object);
void portCallback (SUIT_object object);
void alterValueCallback (SUIT_object object);
void exitProgram (SUIT_object object);

/*--------------------------------------------------------------------------*/

/* Global variables */

static int GsampleRate;
static float Gduration;  /* overall duration, secs */
static float GframeDuration;

static float Gstart,Gend; /* in secs */
static float Glcursor,Grcursor;
static int GsampleStep,GdisplaySamples,GdisplayFrames;
static int GstartFrame,GendFrame;
static double *GparamX=NULL,*GwaveX=NULL;

/*--------------------------------------------------------------------------*/

/* Global SUIT_objects and memory pointers */

/* Parameter Manipulation Area - PMW */
static SUIT_object paramDisplay=NULL;
/* an array of x-coordinates for displaying parameters */

/* volume and output selector */
static unsigned outputPort
#ifdef SUNOS413
=SPEAKER
#endif
;

/* other bits and bobs */
static char *currentFilename="unnamed\0";
static int hasBeenEdited=FALSE;

/* general message assembly space */
static char workSpace[1000];

/*--------------------------------------------------------------------------*/

/* A list of colours to be given to the varied parameters */

char *variedParamColours[]={"blue","red","yellow","green","hotpink",NULL};

/*--------------------------------------------------------------------------*/

#define GREYSCALES 16
static unsigned short rgbValues[GREYSCALES]=
{250,248,244,238,230,220,208,194,178,160,140,118,94,68,40,10};
static int colourIndex[GREYSCALES];

/*--------------------------------------------------------------------------*/

/* TOP AND MIDDLE BAR OBJECTS */

struct cbInfo {
  char *name;
  char *text;
  char *nameOfParent;
  int type;
  SUIT_object object;
  SUIT_callbackFunctionPtr callback;
  int callbackArgument;
} commandBarInfo [] = {
  {"Redraw",NULL,NULL,BUTTON,NULL,redrawCallback,0},

  /*
    LOAD/SAVE COMMANDS
    */
  {"topBar",NULL,NULL,STACKER,NULL,NULL,0},
  {"newButton","New","topBar",BUTTON,NULL,new,0},
  {"loadMenu","Load","topBar",PULLDOWNMENU,NULL,NULL,0},
  {"saveMenu","Save","topBar",PULLDOWNMENU,NULL,NULL,0},
#ifdef SFS
  {"loadSfsFile",".sfs","loadMenu",BUTTON,NULL,loadAudioFile,SFSFILE},
  {"saveSfsFile",".sfs","saveMenu",BUTTON,NULL,saveSfsFile,0},
#endif
  {"loadWavFile",".wav","loadMenu",BUTTON,NULL,loadAudioFile,WAVFILE},
/*  {"loadAuFile",".au","loadMenu",BUTTON,NULL,loadAudioFile,AUFILE},*/

  {"loadParFile",".par","loadMenu",BUTTON,NULL,loadParFile,0},
  {"saveParFile",".par","saveMenu",BUTTON,NULL,saveParFile,0},
  {"exitButton","Exit","topBar",BUTTON,NULL,exitProgram,0},
  /*
    POSITION FEEDBACK
    */
  {"position",NULL,NULL,STACKER,NULL,NULL,0},
  {PARAMNAME,"____","position",MYLABEL,NULL,NULL,0},
  {TIME,"____","position",MYLABEL,NULL,NULL,0},
  {PARAMETERVALUE,"____","position",MYLABEL,NULL,NULL,0},
  {PARAMDESC,"  ",NULL,MYLABEL,NULL,NULL,0},
  
  /* RECALL COMMANDS */
  {"memoriseRecall",NULL,NULL,STACKER,NULL,NULL,0},
  {"recall","Recall","memoriseRecall",BUTTON,NULL,recallParamCallback,0},
  {"hideAll","Hide All","memoriseRecall",BUTTON,NULL,hideAllCallback,0},
  
  /* ARROW BUTTONS */
  {"bigIncrease",NULL,NULL,ARROWBUTTON,NULL,alterValueCallback,10},
  {"smallIncrease",NULL,NULL,ARROWBUTTON,NULL,alterValueCallback,1},
  {"bigDecrease",NULL,NULL,ARROWBUTTON,NULL,alterValueCallback,-10},
  {"smallDecrease",NULL,NULL,ARROWBUTTON,NULL,alterValueCallback,-1},
  
  /* ZOOM WINDOW BUTTONS */ 
  {"zoomStacker",NULL,NULL,STACKER,NULL,NULL,0},
  {"Zoom",NULL,"zoomStacker",BUTTON,NULL,zoomCallback,0},
  {"UnZoom",NULL,"zoomStacker",BUTTON,NULL,unZoomCallback,0},
  
  {"Calc FX",NULL,NULL,BUTTON,NULL,fxCallback,0},
  
  /* RESCALE WAVEFORM BUTTONS */ 
  {"rescaleStacker",NULL,NULL,STACKER,NULL,NULL,0},
  {"increaseW",NULL,"rescaleStacker",ARROWBUTTON,NULL,rescaleCallback,1},
  {"ORIG",NULL,"rescaleStacker",BUTTON,NULL,rescaleCallback,0},
  {"decreaseW",NULL,"rescaleStacker",ARROWBUTTON,NULL,rescaleCallback,-1},
  
  /* SYNTHESIS BUTTON */
  {"synthesise","Synthesise",NULL,BUTTON,NULL,synthesiseCallback,0},


  /* spectra */
  {"normalSpectrum","Spectrum",NULL,BUTTON,NULL,spectrumCallback,0},
  {"smoothedSpectrum","Smoothed",NULL,BUTTON,NULL,spectrumCallback,1},

#ifdef SUNOS413
  /* SIGNAL OUTPUT SWITCHES */
  {"Signaloutput",NULL,NULL,STACKER,NULL,0},
  {"speaker","Speaker","signalOutput",ONOFFSWITCH,NULL,portCallback,SPEAKER},
  {"phones","Headphones","signalOutput",ONOFFSWITCH,NULL,portCallback,PHONES},
  {"line","Line","signalOutput",ONOFFSWITCH,NULL,portCallback,LINE},
  {"release","Release","signalOutput",BUTTON,NULL,closeAudioDevice,0},
#endif
  
  {NULL}, };

/*--------------------------------------------------------------------------*/

#ifdef KLATT88

/* Parameter Number and Location Constants */
#define DU 0
#define UI 1
#define SR 2

/* FIXED PARAMETER INFORMATION */

struct fpInfo {
  int groupNo;
  char *name;
  int Minimum;
  int Default;
  int Maximum;
  char *description;
  SUIT_object button;
  int currentValue;
} fixedParamInfo[] = {
  {1,"DU",30,500,5000,"Duration of the utterance, msec "},
  {1,"UI",1,10,20,"Update interval for parameter reset, msec "},
  {1,"SR",5000,16000,24000,"Output sampling rate, samples/sec "},
  {2,"NF",1,8,8,"Number of formants cascade branch "},
  {2,"SS",1,2,3,"Source switch (1=impulse, 2=natural, 3=LF model) "},
  {2,"RS",1,8,8191,"Random seed (initial value of random # generator) "},
  {3,"SB",0,1,1,"Same noise burst, reset RS if AF=AH=0, 0=no,1=yes "},
  {3,"CP",0,0,1,"0=Cascade, 1=Parallel tract excitation by AV "},
  {3,"OS",0,0,20,"Output selector (0=normal,1=voicing source,...) "},
  {4,"GV",0,60,80,"Overall gain scale factor for AV, dB "},
  {4,"GH",0,60,80,"Overall gain scale factor for AH, dB "},
  {4,"GF",0,60,80,"Overall gain scale factor for AF, dB "},
  {0,NULL},
};
#define NOFIXEDPARAMS ((sizeof(fixedParamInfo)/sizeof(struct fpInfo))-1)

/* VARIED PARAMETER INFORMATION */

#define FX 0
#define NEXTFX 48

struct vpInfo {
  int groupNo;
  char *name;
  int Minimum;
  int Default;
  int Maximum;
  int range;
  char *description;
  GP_color colour;
  SUIT_object button;
  int noFrames;
  double *data;
  double *memorisedData;
  char state;
  char drawn;
} variedParamInfo[] = {
  /* Note that FX is Hz, not in 1/10 Hz, rectified just prior to synthesis */
  {1,"FX",0,100,500,500,"Fundamental Freq., Hz"},
  {9,"AV",0,60,80,100,"Amp. of voicing, dB"},
  {1,"OQ",10,50,99,100,"Open quotient (voicing open-time/period), %"},
  {1,"SQ",100,200,500,500,"Speed quotient (rise/fall time, LF model), %"},
  {1,"TL",0,0,41,100,"Extra tilt of voicing spectrum, dB down at 3 kHz"},
  {1,"FL",0,0,100,100,"Flutter (random fluct in FX), % of maximum"},
  {1,"DI",0,0,100,100,"Diplophonia (alt periods closer), % of max"},
  {9,"AH",0,0,80,100,"Amp. of aspiration, dB"},
  {9,"AF",0,0,80,100,"Amp. of frication, dB"},
  {2,"F1",180,500,1300,8000,"Freq. of F1, Hz"},
  {3,"B1",30,60,1000,500,"BW of F1, Hz"},
  {10,"DF1",0,0,100,5000,"Change in F1 during open portion of period, Hz"},
  {10,"DB1",0,0,400,400,"Change in B1 during open portion of period, Hz"},
  {2,"F2",550,1500,3000,8000,"Freq. of F2, Hz"},
  {3,"B2",40,90,1000,500,"BW of F2, Hz"},
  {2,"F3",1200,2500,4800,8000,"Freq. of F3, Hz"},
  {3,"B3",60,150,1000,500,"BW of F3, Hz"},
  {2,"F4",2400,3250,4990,8000,"Freq. of F4, Hz"},
  {3,"B4",100,200,1000,500,"BW of F4, Hz"},
  {2,"F5",3000,3750,8000,8000,"Freq. of F5, Hz"},
  {3,"B5",100,200,1500,500,"BW of F5, Hz"},
  {2,"F6",3000,4990,4990,8000,"Freq. of F6,in Hz ( if NF=6)"},
  {3,"B6",100,500,4000,500,"BW of F6, Hz ( if NF=6)"},
  {7,"FNP",180,280,500,5000,"Freq. of nasal pole, Hz"},
  {7,"BNP",40,90,1000,500,"BW of nasal pole, Hz"},
  {7,"FNZ",180,280,800,5000,"Freq. of nasal zero, Hz"},
  {7,"BNZ",40,90,1000,500,"BW of nasal zero, Hz"},
  {8,"FTP",300,2150,3000,5000,"Freq. of tracheal pole, Hz"},
  {8,"BTP",40,180,1000,500,"BW of tracheal pole, Hz"},
  {8,"FTZ",300,2150,3000,5000,"Freq. of tracheal zero, Hz"},
  {8,"BTZ",40,180,2000,500,"BW of tracheal zero, Hz"},
  {6,"A2F",0,0,80,100,"Amp. of fric-excited parallel F2, dB"},
  {6,"A3F",0,0,80,100,"Amp. of fric-excited parallel F3, dB"},
  {6,"A4F",0,0,80,100,"Amp. of fric-excited parallel F4, dB"},
  {6,"A5F",0,0,80,100,"Amp. of fric-excited parallel F5, dB"},
  {6,"A6F",0,0,80,100,"Amp. of fric-excited parallel F6, dB"},
  {9,"AB",0,0,80,100,"Amp. of fric-excited parallel bypass path, dB"},
  {5,"B2F",40,250,1000,1000,"BW of fric-excited parallel F2, Hz"},
  {5,"B3F",60,300,1000,1000,"BW of fric-excited parallel F3, Hz"},
  {5,"B4F",100,320,1000,1000,"BW of fric-excited parallel F4, Hz"},
  {5,"B5F",100,360,1500,1000,"BW of fric-excited parallel F5, Hz"},
  {5,"B6F",100,1500,4000,2000,"BW of fric-excited parallel F6, Hz"},
  {7,"ANV",0,0,80,100,"Amp. of voice-excited parallel nasal form., dB"},
  {4,"A1V",0,60,80,100,"Amp. of voice-excited parallel F1, dB"},
  {4,"A2V",0,60,80,100,"Amp. of voice-excited parallel F2, dB"},
  {4,"A3V",0,60,80,100,"Amp. of voice-excited parallel F3, dB"},
  {4,"A4V",0,60,80,100,"Amp. of voice-excited parallel F4, dB"},
  {8,"ATV",0,0,80,100,"Amp. of voice-excited par tracheal formant, dB"},
  {0,NULL},
};
#define NOVARIEDPARAMS ((sizeof(variedParamInfo)/sizeof(struct vpInfo))-1)

/* KLATT88 VARIABLES */

/* Variables are used by the parwav function in parwave.c. 
   When editing, these are not manipulated, but instead filled with the
   appropriate values during synthesis. 
   */

/* Identical to the 12 fixed parameters, except that the UI field is
   expressed in the number of samples per frame instead of ms per frame
   */
int spkrdef[NOFIXEDPARAMS];  

/* Each frame's 48 values are placed in this array prior to calling parwav().
   The FX value for the next frame is placed in the 49th position to allow
   correct interpolation of the FX trajectory for the next frame.
   */
int pars[NOVARIEDPARAMS+1];

long sigmx=0L; /* I haven't got a clue what this means */
int warnsw=1; /* overloads, 1= don't report */
int glotclos=0; /* change open quotient without altering FX */

/* These seem to be used when reporting overload errors to stdout */
char symb[][4]={
  "DU","UI","SR","NF","SS","RS","SB","CP","OS","GV",
  "GH","GF","FX","AV","OQ","SQ","TL","FL","DI","AH",
  "AF","F1","B1","DF1","DB1","F2","B2","F3","B3","F4",
  "B4","F5","B5","F6","B6","FNP","BNP","FNZ","BNZ","FTP",
  "BTP","FTZ","BTZ","A2F","A3F","A4F","A5F","A6F","AB","B2F",
  "B3F","B4F","B5F","B6F","ANV","A1V","A2V","A3V","A4V","ATV"};

/* parwave.c */
extern parwav (short *);

#endif /* #ifdef KLATT88 */

/*--------------------------------------------------------------------------*/

#ifdef KLATT80

#include "parwave.h"

/* KLATT80 VARIABLES */

static int naturalGlottalSamples[]=
{
  -310,-400,530,356,224,89,23,-10,-58,-16,461,599,536,701,770,
  605,497,461,560,404,110,224,131,104,-97,155,278,-154,-1165,
  -598,737,125,-592,41,11,-247,-10,65,92,80,-304,71,167,-1,122,
  233,161,-43,278,479,485,407,266,650,134,80,236,68,260,269,179,
  53,140,275,293,296,104,257,152,311,182,263,245,125,314,140,44,
  203,230,-235,-286,23,107,92,-91,38,464,443,176,98,-784,-2449,
  -1891,-1045,-1600,-1462,-1384,-1261,-949,-730
  };
#define NOGLOTTALSAMPLES ((sizeof(naturalGlottalSamples)/sizeof(int))-1)

klatt_global_t kglobals;

/* Each frame's 40 values are placed in this array prior to calling parwave()*/
klatt_frame_t kframe;

/* FIXED PARAMETER INFORMATION */

/* Parameter Number and Location Constants */
#define DU 0
#define UI 1
#define SR 2

struct fpInfo {
  int groupNo;
  char *name;
  int Minimum;
  int Default;
  int Maximum;
  long *kframePtr;
  char *description;
  SUIT_object button;
  long currentValue;
} fixedParamInfo[] = {
  {1,"DU",30,500,5000,NULL,
     "Duration of the utterance, in msec"},
  {1,"UI",1,10,20,NULL,
     "Update interval for parameter reset, in msec"},
  {1,"SR",5000,9600,24000,&kglobals.samrate,
     "Output sampling rate, samples/sec"},
  {2,"NF",1,5,8,&kglobals.nfcascade,
     "Number of formants in cascade branch"},
  {2,"SS",IMPULSIVE,NATURAL,SAMPLED,&kglobals.glsource,
     "Source (1=impulse, 2=natural, 3=sampled"},
  {2,"CP",CASCADE_PARALLEL,ALL_PARALLEL,ALL_PARALLEL,&kglobals.synthesis_model,
     "Cascade+Parallel/Parallel only switch"},
  {NULL},
};
#define NOFIXEDPARAMS ((sizeof(fixedParamInfo)/sizeof(struct fpInfo))-1)

/* VARIED PARAMETER INFORMATION */

#define FX 0

struct vpInfo {
  int groupNo;
  char *name;
  int Minimum;
  int Default;
  int Maximum;
  int range;
  char *description;
  GP_color colour;
  SUIT_object button;
  int noFrames;
  double *data;
  double *memorisedData;
  char state;
  char drawn;
} variedParamInfo[] = {
  /* Note that FX is Hz, not in 1/10 Hz, rectified just prior to synthesis */
  {1,"FX",0,100,500,500,"Fundamental Freq., Hz"},
  {7,"AV",0,60,80,100,"Amp. of voicing, dB"},
  {2,"F1",180,500,1300,8000,"Freq. of F1, Hz"},
  {3,"B1",30,60,500,500,"BW of F1, Hz"},
  {2,"F2",550,1500,3000,8000,"Freq. of F2, Hz"},
  {3,"B2",40,90,500,500,"BW of F2, Hz"},
  {2,"F3",1200,2500,4800,8000,"Freq. of F3, Hz"},
  {3,"B3",60,150,500,500,"BW of F3, Hz"},
  {2,"F4",2400,3250,4990,8000,"Freq. of F4, Hz"},
  {3,"B4",100,200,500,500,"BW of F4, Hz"},
  {2,"F5",3000,3700,4990,8000,"Freq. of F5, Hz"},
  {3,"B5",100,200,500,500,"BW of F5, Hz"},
  {2,"F6",3000,4990,4990,8000,"Freq. of F6,in Hz (if NF=6)"},
  {3,"B6",100,500,4000,4000,"BW of F6, Hz (if NF=6)"},
  {6,"FNZ",180,280,800,8000,"Freq. of nasal zero, Hz"},
  {6,"BNZ",40,90,1000,500,"BW of nasal zero, Hz"},
  {6,"FNP",180,280,500,8000,"Freq. of nasal pole, Hz"},
  {6,"BNP",40,90,1000,500,"BW of nasal pole, Hz"},
  {7,"ASP",0,0,80,100,"Amp. of aspiration, dB"},
  {1,"KO",10,40,65,100,"# of samples in open period, 10-65"},
  {1,"AT",0,0,80,100,"Breathiness voicing, 0-80"},
  {1,"TL",0,0,24,100,"Voicing spectral tilt in dB, 0-24"},
  {7,"AF",0,0,80,100,"Amp. of frication, dB"},
  {1,"SK",0,0,40,100,"Skewness of alternate periods, 0 to 40 samples"},
  {4,"A1",0,60,80,100,"Amp. of voice-excited parallel F1, dB"},
  {5,"B1P",30,60,500,500,"BW of fric-excited parallel F1, Hz"},
  {4,"A2",0,60,80,100,"Amp. of voice-excited parallel F2, dB"},
  {5,"B2P",40,90,500,500,"BW of fric-excited parallel F2, Hz"},
  {4,"A3",0,60,80,100,"Amp. of voice-excited parallel F3, dB"},
  {5,"B3P",60,150,500,500,"BW of fric-excited parallel F3, Hz"},
  {4,"A4",0,60,80,100,"Amp. of voice-excited parallel F4, dB"},
  {5,"B4P",100,200,500,500,"BW of fric-excited parallel F4, Hz"},
  {4,"A5",0,60,80,100,"Amp. of voice-excited parallel F4, dB"},
  {5,"B5P",100,200,500,500,"BW of fric-excited parallel F5, Hz"},
  {4,"A6",0,60,80,100,"Amp. of voice-excited parallel F4, dB"},
  {5,"B6P",100,500,4000,4000,"BW of fric-excited parallel F6, Hz"},
  {6,"ANP",0,0,80,100,"Amp. of voice-excited parallel nasal form., dB"},
  {7,"AFP",0,0,80,100,"Amp. of fric-excited parallel bypass path, dB"},
  {7,"AVP",0,60,80,100,"Amp. of voicing, parallel, dB"},
  {7,"GA",0,60,60,100,"Overall gain, 60 dB is unity,"},
  {0,NULL},
};
#define NOVARIEDPARAMS ((sizeof(variedParamInfo)/sizeof(struct vpInfo))-1)

#endif  /* #ifdef KLATT80 */

/*--------------------------------------------------------------------------*/

/* Create Screen Objects */

void createObjects ()
     
{
  int number;
  char **colourPtr,*gainControlName,*timeDisplayName,barName[10];
  struct cbInfo *cbi;
  struct vpInfo *vpi;
  struct fpInfo *fpi;
  struct waveInfo *wpi;
  SUIT_object parent;
  /* As many nulls as there are bars of parameters */
  SUIT_object fbarObj[]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
			 NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  SUIT_object vbarObj[]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
			 NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  SUIT_window window;
  
  /* the two waveforms displays */
  for (number=ORIG,colourPtr=variedParamColours;
       number<=SYNTH;
       number++,colourPtr++)
    {
      wpi=&wfi[number];
      wpi->display=SUIT_createObject(wpi->name,"waveform Display");
      /* specify the bounds of the waveform windowx */
      window=SUIT_defWindow(0.0,-1.0,1.0,1.0);
      SUIT_setWindow(wpi->display,WINDOW,window);
      SUIT_setWindow(wpi->display,DEFAULTWINDOW,window);
      SUIT_addDisplayToObject(wpi->display,"standard",
			      waveformHitProc,waveformPaintProc);
      
      /* set up some default values */
      wpi->used=FALSE;
      wpi->waveS=NULL;
      wpi->waveD=NULL;
      wpi->specSize=0;
      wpi->spectrum=NULL;
      wpi->specX=NULL;
      wpi->sgram=NULL;
      wpi->sgramX=NULL;
      wpi->sgramY=NULL;
      wpi->greyScales=NULL;
      
      /* associate a colour with this waveform */
      /* loop if reached the end of the list */
      if (*colourPtr==NULL) colourPtr=variedParamColours;
      wpi->colour=GP_defColor(*colourPtr,TRUE);
      
      SUIT_setInteger(wpi->display,NUMBER,number);
      SUIT_makePropertyTemporary(wpi->display,NUMBER,OBJECT);
      
      /* create bounded value */
      if ((gainControlName=malloc(strlen(wpi->name) + strlen("Gain") 
				  + 2))!=NULL)
	{
	  sprintf(gainControlName,"%sGain",wpi->name);
	  wpi->gainControl=SUIT_createBoundedValue(gainControlName,NULL);
	  free(gainControlName);
	}
      
      /* add time display */
      if ((timeDisplayName=malloc(strlen(wpi->name) + strlen("startTime") 
				  + 2))!=NULL)
	{
	  /*
	  sprintf(timeDisplayName,"%sStartTime",wpi->name);
	  wpi->startTime=SUIT_createLabel(timeDisplayName);
	  SUIT_makePropertyTemporary(wpi->startTime,LABEL,OBJECT);
	  SUIT_setText(wpi->startTime,LABEL,"");
	  
	  sprintf(timeDisplayName,"%sEndTime",wpi->name);
	  wpi->endTime=SUIT_createLabel(timeDisplayName);
	  SUIT_makePropertyTemporary(wpi->endTime,LABEL,OBJECT);
	  SUIT_setText(wpi->endTime,LABEL,"");
	  */

	  sprintf(timeDisplayName,"%sFileName",wpi->name);
	  wpi->fileName=SUIT_createLabel(timeDisplayName);
	  free(timeDisplayName);
	  SUIT_makePropertyTemporary(wpi->fileName,LABEL,OBJECT);
	  SUIT_setText(wpi->fileName,LABEL,"");

	}
    }
  
 /* now parameter display object */
  paramDisplay=SUIT_createObject("pd","Parameter Display Object");
  SUIT_addDisplayToObject(paramDisplay,"standard",
			  parameterHitProc,parameterPaintProc);
  
  /* specify the bounds of the parameter window */
  window=SUIT_defWindow(0.0,0.0,1.0,1.0);
  SUIT_setWindow(paramDisplay,DEFAULTWINDOW,window);
  SUIT_setWindow(paramDisplay,WINDOW,window);
  
  SUIT_setInteger(paramDisplay,EDITEDPARAM,-1);
  SUIT_setInteger(paramDisplay,DISPLAYMODE,PARAMETERS);
  SUIT_makePropertyTemporary(paramDisplay,EDITEDPARAM,OBJECT);

  SUIT_setInteger(paramDisplay,"SGRAM",FALSE);
  SUIT_makePropertyTemporary(paramDisplay,SGRAM,OBJECT);
  
  
  /* create each of the objects in the top and middle command bars */
  for (cbi=commandBarInfo,number=0;
       cbi->name!=NULL;
       cbi++,number++)
    {
      switch(cbi->type)
	{
	case STACKER:
	  cbi->object=SUIT_createStacker(cbi->name);
	  SUIT_setEnumString(cbi->object,ACTIVE_DISPLAY,
			     "horizontal stacking");
	  break;
	  
	case PULLDOWNMENU:
	  cbi->object=SUIT_createPullDownMenu(cbi->name);
	  if (cbi->text!=NULL)
	    SUIT_setText(cbi->object,LABEL,cbi->text);
	  SUIT_setFont(cbi->object,FONT,GP_defFont("times","normal",18));
	  break;
	  
	case BUTTON:
	  cbi->object=SUIT_createButton(cbi->name,cbi->callback);
	  if (cbi->text!=NULL)
	    SUIT_setText(cbi->object,LABEL,cbi->text);
	  SUIT_setFont(cbi->object,FONT,GP_defFont("times","normal",18));
	  break;
	  
	case ARROWBUTTON:
	  cbi->object=SUIT_createArrowButton(cbi->name,cbi->callback);
	  break;
	  
	case ONOFFSWITCH:
	  cbi->object=SUIT_createOnOffSwitch(cbi->name,cbi->callback);
	  if (cbi->text!=NULL)
	    SUIT_setText(cbi->object,LABEL,cbi->text);
	  SUIT_setFont(cbi->object,FONT,GP_defFont("times","normal",12));
	  break;
	  
	case MYLABEL:
	  cbi->object=SUIT_createLabel(cbi->name);
	  if (cbi->text!=NULL)
	    SUIT_setText(cbi->object,LABEL,cbi->text);
	  SUIT_setFont(cbi->object,FONT,GP_defFont("times","normal",18));
	  /* prevent dynamic resizing depending on text */
	  SUIT_setBoolean(cbi->object,SHRINK_TO_FIT,FALSE);
	  /* Another necessary association */
	  SUIT_setObject(paramDisplay,cbi->name,cbi->object);
	  /* Don't save values shown */
	  SUIT_makePropertyTemporary(cbi->object,LABEL,OBJECT);
	  break;
	  
	}
      /* add to parent if necessary */
      if (cbi->nameOfParent!=NULL)
	{
	  parent=SUIT_name(cbi->nameOfParent);
	  SUIT_addChildToObject(parent,cbi->object);
	}
      
      /* associate callback parameter with each object - not used by many*/
      SUIT_setInteger(cbi->object,CALLBACKARGUMENT,cbi->callbackArgument);
      SUIT_makePropertyTemporary(cbi->object,CALLBACKARGUMENT,OBJECT);
      
      /* associate parameter display object with each object */
      SUIT_setObject(cbi->object,DISPLAY,paramDisplay);
    }
  
  /* create each of the time-varying parameter buttons, add to stacker */
  for (vpi=variedParamInfo+NOVARIEDPARAMS-1,
       colourPtr=variedParamColours,number=NOVARIEDPARAMS-1;
       vpi>=variedParamInfo;
       vpi--,colourPtr++,number--)
    {
      /* make a parameter bar if one doesn't exist */
      if (vbarObj[vpi->groupNo]==NULL)
	{
	  sprintf(barName,"vbar%d",vpi->groupNo);
	  vbarObj[vpi->groupNo]=SUIT_createStacker(barName);
	  SUIT_setEnumString(vbarObj[vpi->groupNo],
			     ACTIVE_DISPLAY,"vertical stacking");
	}
      
      /* create a button for this parameter */
      vpi->button=SUIT_createButton(vpi->name,variedParamButtonCallback);
      SUIT_setText(vpi->button,LABEL,vpi->name);
      SUIT_setFont(vpi->button,FONT,GP_defFont("times","normal",12));
      
      /* include object number for referencing information */
      SUIT_setInteger(vpi->button,NUMBER,number);
      SUIT_makePropertyTemporary(vpi->button,NUMBER,OBJECT);
      
      /* associate the display object with the button object */
      SUIT_setObject(vpi->button,DISPLAY,paramDisplay);
      
      /* set this button's initial state and colour */
      vpi->state=NOTDISPLAYED;
      SUIT_setColor(vpi->button,BACKGROUND_COLOR,GP_defColor("gray",TRUE));
      SUIT_setColor(vpi->button,FOREGROUND_COLOR,GP_defColor("black",TRUE));
      SUIT_makePropertyTemporary(vpi->button,BACKGROUND_COLOR,OBJECT);
      SUIT_makePropertyTemporary(vpi->button,FOREGROUND_COLOR,OBJECT);
      
      /* associate a colour with this parameter */
      /* loop if reached the end of the list */
      if (*colourPtr==NULL) colourPtr=variedParamColours;
      vpi->colour=GP_defColor(*colourPtr,TRUE);
      
      /* add it to the stacker bar */
      SUIT_addChildToObject(vbarObj[vpi->groupNo],vpi->button);
      
      /* initialise the data segments to null */
      vpi->noFrames=0;
      vpi->data=NULL;
      vpi->memorisedData=NULL;
    }
  
  /* create each of the fixed parameters */
  for (fpi=fixedParamInfo,number=0;
       fpi->name!=NULL;
       fpi++,number++)
    {
      fpi->currentValue=fpi->Default;
      
      if (fbarObj[fpi->groupNo]==NULL)
	{
	  sprintf(barName,"fbar%d",fpi->groupNo);
	  fbarObj[fpi->groupNo]=SUIT_createStacker(barName);
	  SUIT_setEnumString(fbarObj[fpi->groupNo],
			     ACTIVE_DISPLAY,"horizontal stacking");
	}
      
      /* create a button for this parameter and add it to the stacker */
      fpi->button=SUIT_createButton(fpi->name,fixedParamButtonCallback);
      SUIT_addChildToObject(fbarObj[fpi->groupNo],fpi->button);
      
      /* include object number for referencing information */
      SUIT_setInteger(fpi->button,NUMBER,number);
      SUIT_makePropertyTemporary(fpi->button,NUMBER,OBJECT);
    }
}

/*-------------------------------------------------------------------------*/

void initialise (int noFrames, int updateInterval, int sampleRate,
		 int reset, int synthesise, int display)
     
{
  int frame,sample,noSamples,sampleNo;
  double defaultValue,*dataPtr,scale;
  struct vpInfo *vpi;
  struct fpInfo *fpi;
  struct waveInfo *wpi;
  char buttonLabel[20];
  
#ifdef KLATT80
  /* fill in a few default values */
  kglobals.sample_factor = 1.0;
  kglobals.f0_flutter = 0;
  kglobals.natural_samples = naturalGlottalSamples;
  kglobals.num_samples = NOGLOTTALSAMPLES;
#endif
  
  wfi[ORIG].recalcSpec=TRUE;
  wfi[ORIG].recalcSgram=TRUE;
  wfi[ORIG].recalcGreyScales=TRUE;

  wpi=&wfi[SYNTH];
  
  /* calculate the desired duration */
  GframeDuration= (float)updateInterval/1000.;
  Gduration=GframeDuration * (float)noFrames;
  fixedParamInfo[DU].currentValue=(int)(Gduration * 1000.);
  
  GsampleRate=sampleRate;
  noSamples= (int)( Gduration * GsampleRate);
  wpi->used=TRUE;
  
  /* initialise the synthesised waveform display */
  wpi->recalcSpec=TRUE;
  wpi->recalcGreyScales=TRUE;
  wpi->recalcSgram=TRUE;
  
  /* only a certain number of waveform points can be shown, may
     have to show only every second or third sample depending on
     the duration and the sample rate. */
  GsampleStep=1;
  while ( (int)((float)noSamples/(float)GsampleStep) > MAXPOINTS)
    GsampleStep++;
  GdisplaySamples=noSamples/GsampleStep;
  
  /* allocate space for the waveform, and the displayed waveform */
  myAllocate((void *)&wpi->waveS,(int)(1.1*noSamples),SHORT);
  myAllocate((void *)&wpi->waveD,GdisplaySamples,DOUBLE);
  
  /* convert original waveform to doubles if needed */
  if (wfi[ORIG].used)
    {
      myAllocate((void *)&wfi[ORIG].waveD,GdisplaySamples,DOUBLE);
      for (sampleNo=0,sample=0;
	   sampleNo<GdisplaySamples;
	   sampleNo++,sample+=GsampleStep)
	wfi[ORIG].waveD[sampleNo]=(double)wfi[ORIG].waveS[sample]/MAXSAMPLE;
    }
  
  /* calculate the X coordinate values */
  myAllocate((void *)&GwaveX,GdisplaySamples,DOUBLE);
  for (sample=0;sample<GdisplaySamples;sample++)
    GwaveX[sample]=(double)sample/(double)(GdisplaySamples-1);
  
  /* initialise zoom values */
  Glcursor=Gstart=0.;
  Grcursor=Gend=Gduration;
  GstartFrame=0;GendFrame=noFrames;
  
  /* now the parameter display */
  /* calculate an array of x-coordinates for displaying parameters */
  myAllocate((void *)&GparamX,noFrames,DOUBLE);
  for (frame=0;frame<noFrames;frame++)
    GparamX[frame]=(double)frame/(double)(noFrames-1);
  GdisplayFrames=noFrames;
  
  if (display)
    {
      /* set initial display mode */
      SUIT_setInteger(paramDisplay,NOFRAMES,noFrames);
      SUIT_makePropertyTemporary(paramDisplay,NOFRAMES,OBJECT);
      SUIT_setInteger(paramDisplay,DISPLAYMODE,PARAMETERS);
      SUIT_setInteger(paramDisplay,EDITEDPARAM,-1);
    }
  
  for (vpi=variedParamInfo;
       vpi->name!=NULL;
       vpi++)
    {
      /* allocate space for the storage of this parameter's value */
      myAllocate((void *)&vpi->data,noFrames,DOUBLE);
      myAllocate((void *)&vpi->memorisedData,noFrames,DOUBLE);
      
      /* set this button's state and colour */
      vpi->state=NOTDISPLAYED;
      if (display)
	{
	  SUIT_setColor(vpi->button,BACKGROUND_COLOR,GP_defColor("gray",TRUE));
	  SUIT_setColor(vpi->button,FOREGROUND_COLOR,GP_defColor("black",TRUE));
	}
      
      if ((noFrames>vpi->noFrames) || (reset))
	{
	  /* reset hack */
	  if (reset)
	    vpi->noFrames=0;
	  
	  if (vpi->noFrames==0)
	    defaultValue=(double)vpi->Default/(double)vpi->range;
	  else
	    defaultValue=(double)vpi->data[(vpi->noFrames)-1];
	  
	  /* set any new data to this value */
	  if (vpi->noFrames==0) dataPtr=vpi->data;
	  else dataPtr=vpi->data + vpi->noFrames;
	  for (frame=0;
	       frame<(noFrames-vpi->noFrames);
	       frame++,dataPtr++)
	    *dataPtr=defaultValue;
	  
	  /* and any saved data */
	  if (vpi->noFrames==0) dataPtr=vpi->memorisedData;
	  else dataPtr=vpi->memorisedData + vpi->noFrames;
	  for (frame=0;
	       frame<(noFrames-vpi->noFrames);
	       frame++,dataPtr++)
	    *dataPtr=defaultValue;
	}
      vpi->noFrames=noFrames;
    }
  
  /* adjust the formant frequency range to half the sample rate */
  for (vpi=variedParamInfo;
       vpi->name!=NULL;
       vpi++)
    {
      if (vpi->groupNo==2)
	{
	  scale=(double)vpi->range/((double)sampleRate/2.);
	  for (frame=0,dataPtr=vpi->data;
	       frame<vpi->noFrames;
	       frame++,dataPtr++)
	    *dataPtr *=scale;
	  vpi->range =(int)((double)vpi->range/scale);
	}
    }
  
  /* update the values of the fixed parameters */
  for (fpi=fixedParamInfo;
       fpi->name!=NULL;
       fpi++)
    {
      if (reset)
	fpi->currentValue=fpi->Default;
      
      /* fill in the label */
      if (display)
	{
	  SUIT_setFont(fpi->button,FONT,GP_defFont("times","normal",12));
	  sprintf(buttonLabel,"%s : %d",fpi->name,fpi->currentValue);
	  SUIT_setText(fpi->button,LABEL,buttonLabel);
	}
    }
  
  if (display)
    SUIT_redisplayRequired(paramDisplay);
  
  if (synthesise)
    synthesiseCallback(NULL);
  
}

/*--------------------------------------------------------------------------*/

void myAllocate (void **pptr, size_t nelem, int type)
{
  /* hack */
  nelem*=2;
  /* reallocate if already been allocated */
  if (*pptr==NULL)
    {
      switch (type)
	{
	case CHAR:*pptr=(char *)calloc(nelem,sizeof(char));break;
	case SHORT:*pptr=(short *)calloc(nelem,sizeof(short));break;
	case FLOAT:*pptr=(float *)calloc(nelem,sizeof(float));break;
	case DOUBLE:*pptr=(double *)calloc(nelem,sizeof(double));break;
	}
      if (*pptr==NULL) goto crashGracefully;
    }
  else
    {
      switch (type)
	{
	case CHAR:*pptr=(char *)realloc(*pptr,nelem*sizeof(char));break;
	case SHORT:*pptr=(short *)realloc(*pptr,nelem*sizeof(short));break;
	case FLOAT:*pptr=(float *)realloc(*pptr,nelem*sizeof(float));break;
	case DOUBLE:*pptr=(double *)realloc(*pptr,nelem*sizeof(double)); break;
	}
      if (*pptr==NULL) goto crashGracefully;
    }
  return;

 crashGracefully:
  SUIT_inform("KPE has memory problems - please save your parameter file");
  saveParFile(NULL);
  exit(1);
}

/*-------------------------------------------------------------------------*/

void redrawCallback (SUIT_object object)
     
{
  /* redraw absolutely everything */
  SUIT_allObjectsRequireRedisplay(NULL);
  SUIT_performRedisplay();
}

/*-------------------------------------------------------------------------*/

void variedParamButtonCallback (SUIT_object paramButton)
     
{
  int number,currentNumber;
  struct vpInfo *vpi,*currentVpi;
  SUIT_object display;
  
  /* first press in spectra mode changes mode only */
  if (SUIT_getInteger(paramDisplay,DISPLAYMODE)==SPECTRA)
    {
      SUIT_setInteger(paramDisplay,DISPLAYMODE,PARAMETERS);
      return;
    }
  
  /* locate this parameter's data */
  number=SUIT_getInteger(paramButton,NUMBER);
  vpi=&variedParamInfo[number];
  
  display=SUIT_getObject(paramButton,DISPLAY);
  switch (vpi->state)
    {
    case NOTDISPLAYED:
      /* display this button */
      vpi->state=DISPLAYED;
      SUIT_setColor(vpi->button,FOREGROUND_COLOR,vpi->colour);
      SUIT_setColor(vpi->button,BACKGROUND_COLOR,GP_defColor("gray",TRUE));
      SUIT_redisplayRequired(vpi->button);
      
      /* turn any parameter that's being edited off - so that markings on
	 Y-axis disappear - this will occur when the display is repainted
	 */
      for (vpi=variedParamInfo;
	   vpi->name!=NULL;
	   vpi++)
	{
	  if (vpi->state==BEINGEDITED)
	    {
	      vpi->state=DISPLAYED;
	      
	      SUIT_setColor(vpi->button,FOREGROUND_COLOR,
			    vpi->colour);
	      SUIT_setColor(vpi->button,BACKGROUND_COLOR,
			    GP_defColor("gray",TRUE));
	      SUIT_redisplayRequired(vpi->button);
	      /* don't bother looking any further */
	      vpi->drawn=FALSE;
	      break;
	    }
	}
      break;
      
    case DISPLAYED:
      /* edit this parameter */
      
      /* first change status of current edited parameter */
      if ((currentNumber=SUIT_getInteger(display,EDITEDPARAM))>=0)
	{
	  currentVpi=&variedParamInfo[currentNumber];
	  currentVpi->state=DISPLAYED;
	  SUIT_setColor(currentVpi->button,FOREGROUND_COLOR,
			currentVpi->colour);
	  SUIT_setColor(currentVpi->button,BACKGROUND_COLOR,
			GP_defColor("gray",TRUE));
	  SUIT_redisplayRequired(currentVpi->button);
	}
      
      /* now change status of current one */
      vpi->state=BEINGEDITED;
      SUIT_setInteger(display,EDITEDPARAM,number);
      SUIT_setColor(vpi->button,BACKGROUND_COLOR,vpi->colour);
      SUIT_setColor(vpi->button,FOREGROUND_COLOR,GP_defColor("black",TRUE));
      SUIT_redisplayRequired(vpi->button);
      vpi->drawn=FALSE;
      break;
      
    case BEINGEDITED:
      vpi->state=NOTDISPLAYED;
      SUIT_setColor(vpi->button,FOREGROUND_COLOR,GP_defColor("black",TRUE));
      SUIT_setColor(vpi->button,BACKGROUND_COLOR,GP_defColor("gray",TRUE));
      SUIT_redisplayRequired(vpi->button);
      SUIT_setInteger(display,EDITEDPARAM,-1);
      vpi->drawn=FALSE;
      break;
    }

  /* prevent spectrogram being overwritten */
  if (SUIT_getInteger(paramDisplay,SGRAM))
    SUIT_redisplayNotRequired(paramDisplay);
  else
    SUIT_redisplayRequired(paramDisplay);
}

/*-------------------------------------------------------------------------*/

void fixedParamButtonCallback (SUIT_object paramButton)
     
{
  int newValue,newNoFrames,frame,nearestSR=0;
  double *dataPtr,*memDataPtr;
  struct fpInfo *fpi;
  struct vpInfo *vpi;
  char currentValue[10],answer[10],newLabel[20];
  
  fpi=&fixedParamInfo[SUIT_getInteger(paramButton,NUMBER)];
  
  sprintf(workSpace,"%s: %s.  Please enter a new value between %d and %d",
	  fpi->name,fpi->description,fpi->Minimum,fpi->Maximum);
  sprintf(currentValue,"%d",fpi->currentValue);
  
  /* loop until user enters a legal value */
  do
    {
      SUIT_getString(workSpace,currentValue,answer,10);
      newValue=atoi(answer);
    }
  while ((newValue<fpi->Minimum)||(newValue>fpi->Maximum));
  
#ifdef SUNOS413
  if (strcmp(fpi->name,"SR")==0)
    {
      if ((nearestSR=nearestSunSampleRate(newValue))!=newValue)
	{
	  sprintf(workSpace,"Nearest Sun sample rate = %d",nearestSR);
	  SUIT_inform(workSpace);
	  return;
	}

      /* resample original waveform to this new rate */

      resampleWave(fpi->currentValue,
		   newValue,
		   &wfi[ORIG].waveS,
		   (int)(Gduration*(float)GsampleRate));
    }
#endif

  /* set the new value */
  fpi->currentValue=newValue;
  
  /* allocate any extra memory required */
  if ((strcmp(fpi->name,"DU")==0) ||
      (strcmp(fpi->name,"SR")==0))
    {
      initialise((fixedParamInfo[DU].currentValue/
		  fixedParamInfo[UI].currentValue),
		 fixedParamInfo[UI].currentValue,
		 fixedParamInfo[SR].currentValue,
		 FALSE,
		 TRUE,  /* resynthesise */
		 TRUE); /* display */
    }
  else 
    {
      if (strcmp(fpi->name,"UI")==0)
	{
	  GframeDuration=(float)fixedParamInfo[UI].currentValue/1000.;
	  newNoFrames=(int)(Gduration/GframeDuration);
	  /* resample the parameter values */
	  for (vpi=variedParamInfo;
	       vpi->name!=NULL;
	       vpi++)
	    if (vpi->noFrames>0)
	      {
		resample (vpi->noFrames, newNoFrames, &(vpi->data));
		vpi->noFrames=newNoFrames;
		/* copy data to memorised data */
		myAllocate((void *)&vpi->memorisedData,newNoFrames,DOUBLE);
		for (frame=0,dataPtr=vpi->data,memDataPtr=vpi->memorisedData;
		     frame<vpi->noFrames;
		     frame++,dataPtr++,memDataPtr++)
		  *memDataPtr=*dataPtr;
	      }
	  unZoomCallback(NULL);
	  
	  SUIT_setInteger(paramDisplay,NOFRAMES,newNoFrames);
	  SUIT_redisplayRequired(paramDisplay);}
	  
      
      /* update label if not using initialise */
      sprintf(newLabel,"%s : %d",fpi->name,newValue);
      SUIT_setText(paramButton,LABEL,newLabel);
      synthesiseCallback(NULL);
    }
}

/*-------------------------------------------------------------------------*/

void resample (int oldNoFrames, int newNoFrames, double **dataPtr)
     
{
  int newFrame,oldFrameDiv;
  float fraction,oldFrameMod;
  double newValue,*oldData,*newData;
  double lowerOldFrameValue,upperOldFrameValue;
  
  /* allocate some new space into which to resample the data */
  newData=NULL;
  myAllocate((void *)&newData,newNoFrames+10,DOUBLE);
  
  oldData=*dataPtr;
  
  /* go through and resample */
  for (newFrame=0;newFrame<newNoFrames;newFrame++)
    {
      /* express current position as a fraction between 0 and 1 */
      fraction=(float)newFrame/(float)newNoFrames;
      
      /* calculate where this Frame lies in the old data */
      oldFrameDiv=(int)(fraction*(float)oldNoFrames);
      oldFrameMod=(fraction * (float)oldNoFrames) - (float)oldFrameDiv;
      
      /* calculate the value at the new Frame, interpolating */
      lowerOldFrameValue=oldData[oldFrameDiv];
      
      /* avoid interpolating past the end of the old data */
      if (oldFrameDiv>=(oldNoFrames-1))
	upperOldFrameValue=oldData[oldFrameDiv];
      else
	upperOldFrameValue=oldData[oldFrameDiv+1];
      
      newValue= lowerOldFrameValue + 
	(oldFrameMod * (upperOldFrameValue - lowerOldFrameValue));
      
      /*       printf("newFrame %d  oldFrameDiv %d  oldFrameMod %.2f oldValue %.2f newValue %.2f\n",
	       newFrame,oldFrameDiv,oldFrameMod,oldData[oldFrameDiv],newValue);
	       */	     
      
      newData[newFrame]= newValue;
    }
  
  /* swap old and new data, freeing old data */
  *dataPtr=newData;
  /*  free(oldData);*/
}

/*--------------------------------------------------------------------------*/

/* Linestart must be visible to the lineDragCallback function */

void parameterHitProc (SUIT_object display, SUIT_event myEvent)
     
{
  register int bin,spectrum,noBins;
  register double *xPtr,*yPtr,yScale,yOffset;
  register char *greyScalePtr;
  
  struct waveInfo *wpi;
  struct vpInfo *vpi;
  int count,steps,startFrame,endFrame,number,key,writeMode;
  double range,*dataPtr,increment;
  GP_point position;
  double tempy,tempx;
  static struct vpInfo *editedVpi=NULL;
  static int lineBeingDrawn;
  static GP_point lineStart, oldPosition;
  static GP_color parameterColour,backgroundColour;
  
  position=myEvent.worldLocation;
  writeMode=SUIT_getInteger(display,WRITEMODE);
  
  switch(SUIT_getInteger(display,DISPLAYMODE))
    {
    case SPECTRA:
      switch(myEvent.type)
	{
	case MOUSE_DOWN:
	case MOUSE_MOTION:
	  switch (myEvent.button)
	    {
	    case MIDDLE_BUTTON:
	      displayPosition(display, position);
	      break;
	    }
	case KEYSTROKE:
	  SUIT_setInteger(paramDisplay,DISPLAYMODE,PARAMETERS);
	  goto keyStroke;
	}

    case PARAMETERS:
      
      switch(myEvent.type)
	{
	case KEYSTROKE:
	keyStroke:
	  switch (key=myEvent.keyboard)
	    {
	    case ' ':
	      synthesiseCallback(NULL);
	      break;
	      
	      /* Spectrogram */
	    case 'b':
	    case 's':
	    case 'o':
	      
	      if (key=='b')
		{yOffset=0.5;yScale=0.5;}
	      else
		{yOffset=0.;yScale=1.0;}
	      
	      if ((key=='s') || (key=='b'))
		{
		  GP_setCursor(WATCH_CURSOR);
  
		  /* synthetic above */
		  wpi=&wfi[SYNTH];
		  if (wpi->recalcSgram) 
		    calculateGreyscales(wpi,calculateSpectrogram(wpi));
		  
		  GP_setWriteMode(WRITE_REPLACE);
		  noBins=wpi->fftSize/2;
		  for (spectrum=0,greyScalePtr=wpi->greyScales,xPtr=wpi->sgramX;
		       spectrum<NOSPECTRA;
		       spectrum++,xPtr++)
		    {
		      for (bin=0,yPtr=wpi->sgramY;
			   bin<noBins;
			   bin++,greyScalePtr++,yPtr++)
			{
			  SRGP_setColor((int)*greyScalePtr);
			  GP_pointCoord(*xPtr,(yOffset+(yScale* *yPtr)));
			}
		    }
		}
	      
	      if ((key=='o') || (key=='b'))
		{
		  GP_setCursor(WATCH_CURSOR);

		  /* original below */
		  wpi=&wfi[ORIG];
		  if (wpi->recalcSgram)
		    calculateGreyscales(wpi,calculateSpectrogram(wpi));
		  
		  GP_setWriteMode(WRITE_REPLACE);
		  noBins=wpi->fftSize/2;
		  for (spectrum=0,greyScalePtr=wpi->greyScales,xPtr=wpi->sgramX;
		       spectrum<NOSPECTRA;
		       spectrum++,xPtr++)
		    {
		      for (bin=0,yPtr=wpi->sgramY;
			   bin<noBins;
			   bin++,greyScalePtr++,yPtr++)
			{
			  SRGP_setColor((int)*greyScalePtr);
			  GP_pointCoord(*xPtr,(yScale * *yPtr));
			}
		    }
		}
	      SUIT_setInteger(display,SGRAM,TRUE);
	      
	      /* now draw parameters on top */
	      SUIT_setInteger(display,WRITEMODE,WRITE_XOR);
	      GP_setWriteMode(WRITE_XOR);
	      GP_setLineWidth(LINEWIDTH);
	      for (vpi=variedParamInfo;
		   vpi->name!=NULL;
		   vpi++)
		{
		  switch(vpi->state)
		    {
		      
		    case DISPLAYED:
		      GP_setLineStyle(DOTTED);
		      GP_setColor(vpi->colour);
		      GP_polyLineCoord(GdisplayFrames,
				       GparamX,&vpi->data[GstartFrame]);
		      vpi->drawn=TRUE;
		      break;
		      
		    case BEINGEDITED:
		      editedVpi=vpi;
		      GP_setLineStyle(SOLID);
		      GP_setColor(vpi->colour);
		      GP_polyLineCoord(GdisplayFrames,
				       GparamX,&vpi->data[GstartFrame]);
		      vpi->drawn=TRUE;
		      break;
		    }
		}
	      
	      GP_setLineStyle(SOLID);
	      GP_setCursor(STANDARD_CURSOR);
	      SUIT_redisplayNotRequired(display);
	      break;
	    }
	  
	case MOUSE_DOWN:
	  switch (myEvent.button)
	    {
	    case LEFT_BUTTON:
	      /* draw a line between two points */
	      if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
		{
		  editedVpi=&variedParamInfo[number];
		  lineBeingDrawn=TRUE;
		  lineStart=position;
		  oldPosition=lineStart;
		  parameterColour=editedVpi->colour;
		  backgroundColour=SUIT_getColor(display,BACKGROUND_COLOR);
		  hasBeenEdited=TRUE;
		}
	      break;
	      
	    case MIDDLE_BUTTON:
	      if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
		displayPosition(display, position);
	      break;
	    }
	  break;
	  
	case MOUSE_MOTION:
	  switch (myEvent.button)
	    {
	    case LEFT_BUTTON:
	      if (lineBeingDrawn)
		{
		  GP_setWriteMode(writeMode);
		  GP_setColor(parameterColour);
		  GP_setLineStyle(SOLID);
		  GP_setLineWidth(LINEWIDTH);
		  
		  /* remove previous line */
		  if (writeMode!=WRITE_XOR)
		    GP_setColor(SUIT_getColor(display,BACKGROUND_COLOR));
		  GP_lineCoord(lineStart.x,lineStart.y,
			       oldPosition.x,oldPosition.y);
		  
		  /* draw in the new one */
		  /* remove previous line */
		  GP_setColor(parameterColour);
		  GP_lineCoord(lineStart.x,lineStart.y,
			       position.x,position.y);
		  oldPosition=position;
		  
		  displayPosition(display, position);
		  GP_setLineWidth(1);
		}
	      break;
	      
	    case MIDDLE_BUTTON:
	      if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
		displayPosition(display, position);
	      break;
	    }
	  break;
	  
	case MOUSE_UP:
	  switch (myEvent.button)
	    {
	    case LEFT_BUTTON:
	      if (lineBeingDrawn)
		{
		  /* remove the line */
		  GP_setWriteMode(writeMode);
		  if (writeMode==WRITE_XOR)
		    GP_setColor(parameterColour);
		  else
		    GP_setColor(SUIT_getColor(display,BACKGROUND_COLOR));
		  
		  GP_setLineWidth(LINEWIDTH);
		  GP_lineCoord(lineStart.x,lineStart.y,
			       oldPosition.x,oldPosition.y);
		  GP_setLineWidth(1);
		  
		  /* now interpolate data */
		  
		  /* first check sign */
		  if (lineStart.x>position.x)
		    {
		      tempx=position.x;tempy=position.y;
		      position.x=lineStart.x;position.y=lineStart.y;
		      lineStart.x=tempx;lineStart.y=tempy;
		    }
		  
		  
		  /* amend trajectory */
		  startFrame=GstartFrame + 
		    (int)((lineStart.x * GdisplayFrames)+.5);
		  endFrame=GstartFrame + 
		    (int)((position.x * GdisplayFrames)+.5);
		  
		  steps=(endFrame-startFrame);
		  range=position.y-lineStart.y;
		  if (steps>0)
		    {
		      if (editedVpi->drawn)
			{
			  /* remove the old parameter track */
			  GP_setWriteMode(writeMode);
			  if (writeMode==WRITE_XOR)
			    GP_setColor(parameterColour);
			  else
			    GP_setColor(SUIT_getColor(display,BACKGROUND_COLOR));
			  GP_setLineWidth(LINEWIDTH);
			  GP_polyLineCoord(GdisplayFrames,
				       GparamX,
				       &editedVpi->data[GstartFrame]);
			}

		      /* amend values */
		      increment=(range/(double)steps);
		      for (count=0,dataPtr=&(editedVpi->data[startFrame]);
			   count<=steps;
			   count++,dataPtr++)
			*dataPtr=lineStart.y + (increment*count);
		      
		      /* draw in the new track */
		      GP_setLineWidth(LINEWIDTH);
		      GP_setWriteMode(writeMode);
		      GP_setColor(parameterColour);
		      GP_polyLineCoord(GdisplayFrames,
				       GparamX,
				       &editedVpi->data[GstartFrame]);
		      GP_setLineWidth(1);
		      editedVpi->drawn=TRUE;
		    }
		}
	      lineBeingDrawn=FALSE;
	      break;
	    }
	  break;
	  
	case CLICK:
	  switch (myEvent.button)
	    {		    
	    case LEFT_BUTTON:
	      displayPosition(display, position);
	      break;
	    case RIGHT_BUTTON:
	    case MIDDLE_BUTTON:
	      SUIT_redisplayRequired(paramDisplay);
	      break;
	    }
	  break;
	  
	}
      break;
      
    }
  GP_setLineWidth(1);
  GP_setWriteMode(WRITE_REPLACE);
}

/*-------------------------------------------------------------------------*/

void displayPosition (SUIT_object display, GP_point position)
     
{
  int frameNo,number,noFrames;
  char timeText[10],paramValueText[10];
  char freqText[10],magText[10];
  struct vpInfo *vpi;
  
  /* round to the nearest frame number */
  noFrames=SUIT_getInteger(display,NOFRAMES);
  frameNo=(int)(((double)noFrames * position.x) + .5);
  
  switch(SUIT_getInteger(display,DISPLAYMODE))
    {
    case PARAMETERS:
      sprintf(timeText,"%.3fs",
	      ((float)fixedParamInfo[UI].currentValue/1000.)*(float)frameNo);
      SUIT_setText(SUIT_getObject(display,TIME),LABEL,timeText);
      
      if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
	{
	  vpi=&variedParamInfo[number];
	  sprintf(paramValueText,"%4d",(int)(position.y * vpi->range));
	  SUIT_setText(SUIT_getObject(display,PARAMETERVALUE),LABEL,paramValueText);
	}
      break;
      
    case SPECTRA:
      sprintf(freqText,"%d",(int)(position.x * ((float)GsampleRate/2.0)));
      sprintf(magText,"%.1fdB",position.y * SPECRANGE);
      SUIT_setText(SUIT_getObject(display,PARAMNAME),LABEL,"DFT");
      SUIT_setText(SUIT_getObject(display,TIME),LABEL,freqText);
      SUIT_setText(SUIT_getObject(display,PARAMETERVALUE),LABEL,magText);
      break;
    }
}

/*-------------------------------------------------------------------------*/

void parameterPaintProc (SUIT_object display)
     
{
  int number,frame,noFrames,frameStep;
  int frequencyStep,frequency,sgram;
  double x,y,yStep;
  double lposn,rposn;
  char text[10];
  GP_point position;
  SUIT_window window;
  struct vpInfo *vpi,*editedvpi;
  struct waveInfo *wpi;
  
  editedvpi=NULL;
  SUIT_setInteger(display,"SGRAM",FALSE);
  
  switch (SUIT_getInteger(display,DISPLAYMODE))
    {
    case PARAMETERS:

      GP_setColor(GP_defColor("black",TRUE));
      GP_setLineStyle(SOLID);
      GP_setLineWidth(1);
      lposn=(Glcursor-Gstart)/(Gend-Gstart);
      GP_lineCoord(lposn,0.,lposn,1.0);
      rposn=(Grcursor-Gstart)/(Gend-Gstart);
      GP_lineCoord(rposn,0.,rposn,1.0);

      GP_setLineWidth(LINEWIDTH);
      window=SUIT_getWindow(display,WINDOW);
      GP_drawRectangle(window);
      
      SUIT_setInteger(display,WRITEMODE,WRITE_REPLACE);
      GP_setWriteMode(WRITE_REPLACE);

      for (vpi=variedParamInfo;
	   vpi->name!=NULL;
	   vpi++)
	{
	  switch(vpi->state)
	    {
	      
	    case DISPLAYED:
	      GP_setLineStyle(DOTTED);
	      GP_setColor(vpi->colour);
	      GP_polyLineCoord(GdisplayFrames,
			       GparamX,&vpi->data[GstartFrame]);
	      vpi->drawn=TRUE;
	      break;
	      
	    case BEINGEDITED:
	      editedvpi=vpi;
	      GP_setLineStyle(SOLID);
	      GP_setColor(vpi->colour);
	      GP_polyLineCoord(GdisplayFrames,
			       GparamX,&vpi->data[GstartFrame]);
	      vpi->drawn=TRUE;
	      break;
	    }
	}
      
      GP_setColor(GP_defColor("black",TRUE));
      GP_setWriteMode(WRITE_REPLACE);
      
      /* ANNOTATE X-AXIS */
      
      noFrames=SUIT_getInteger(display,NOFRAMES);
      /* frame number markings */
      if (noFrames<=100)
	frameStep=5;
      else if (noFrames<=200)
	frameStep=10;
      else if (noFrames<=500)
	frameStep=25;
      else if (noFrames<=1000)
	frameStep=50;
      else if (noFrames<=5000)
	frameStep=500;
      for (frame=frameStep;frame<noFrames;frame+=frameStep)
	{
	  x=(double)frame/(double)noFrames;
	  
	  /* put a vertical tick */
	  GP_lineCoord(x,1.0,x,0.98);
	  GP_lineCoord(x,0.0,x,0.02);
	  
	  /* draw the frame number */
	  position.x=x+.005;
	  position.y=0.975;
	  sprintf(text,"%d",frame);
	  GP_text(position,text);
	  
	  /* draw the time position */
	  position.x=x+.005;
	  position.y=0.005;
	  sprintf(text,"%ldms",frame*fixedParamInfo[UI].currentValue);
	  GP_text(position,text);
	}
      
      /* ANNOTATE Y-AXIS */
      
      /* only mark if a parameter is being edited */
      if (editedvpi!=NULL)
	{
	  /* now mark the y-axis */
	  yStep=0.1;
	  for (y=yStep;y<0.99;y+=yStep)
	    {
	      /* first the ticks */
	      GP_lineCoord(0.0,y,0.005,y);
	      GP_lineCoord(1.0,y,0.995,y);
	      
	      sprintf(text,"%d",(int)((y*(double)editedvpi->range)+.5));
	      position.x=0.01;
	      position.y=y-0.01;
	      GP_text(position,text);
	      position.x=0.96;
	      GP_text(position,text);
	    }
	}
      
      /* ECHO SYNTHESISED WAVEFORM MARKERS */
      wpi=&wfi[SYNTH];
      
      /* DISPLAY PARAMETER NAME */
      
      if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
	{
	  vpi=&variedParamInfo[number];
	  sprintf(text,"%3.3s",vpi->name);
	  SUIT_setText(SUIT_getObject(display,PARAMNAME),LABEL,text);
	  SUIT_setText(SUIT_getObject(display,PARAMDESC),LABEL,vpi->description);
	}
      else
	{
	  SUIT_setText(SUIT_getObject(display,PARAMNAME),LABEL,"___");
	  SUIT_setText(SUIT_getObject(display,PARAMDESC),LABEL," ");
	}
      break;
      
      
    case SPECTRA:
      
      for (wpi=wfi;
	   wpi->name!=NULL;
	   wpi++)
	{
	  if (wpi->defineWindow && (wpi->specSize >0) )
	    {
	      /* define the new window */
	      GP_setLineWidth(1);
	      GP_setColor(GP_defColor("black",TRUE));
/*	      GP_drawRectangle(window);*/
	    }
	  if (wpi->specSize>0)
	    {
	      GP_setLineWidth(1);
	      GP_setLineStyle(SOLID);
	      GP_setColor(wpi->colour);
	      GP_polyLineCoord(wpi->specSize,
			       wpi->specX,wpi->spectrum);
	    }
	}
      
      GP_setColor(GP_defColor("black",TRUE));
      
      /* ANNOTATE X-AXIS */
      
      if (GsampleRate>10000)
	frequencyStep=1000;
      else
	frequencyStep=500;
      
      for (frequency=frequencyStep;
	   frequency<GsampleRate/2;
	   frequency+=frequencyStep)
	{
	  x=(double)frequency/((double)GsampleRate/2.0);
	  
	  /* put a vertical tick */
	  GP_lineCoord(x,1.0,x,0.98);
	  GP_lineCoord(x,0.0,x,0.02);
	  
	  /* draw the frequency*/
	  position.x=x+.005;
	  position.y=0.975;
	  sprintf(text,"%.1fk",frequency/1000.);
	  GP_text(position,text);
	  position.y=0.005;
	  GP_text(position,text);
	}
      
      /* ANNOTATE Y-AXIS */
      
      /* now mark the y-axis */
      yStep=0.1;
      for (y=yStep;y<0.99;y+=yStep)
	{
	  /* first the ticks */
	  GP_lineCoord(0.0,y,0.005,y);
	  GP_lineCoord(1.0,y,0.995,y);
	  
	  sprintf(text,"%d",(int)((y*SPECRANGE)+.5));
	  position.x=0.01;
	  position.y=y-0.01;
	  GP_text(position,text);
	  position.x=0.96;
	  GP_text(position,text);
	}
      
      break;
      
    }
  
  /* undo any changes */
  GP_setWriteMode(WRITE_REPLACE);
  GP_setLineWidth(1);
}

/*-------------------------------------------------------------------------*/

float calculateSpectrogram (struct waveInfo *wpi)
     
{
  register int sample;
  register float *fftPtr,*windowPtr,xsum;
  register short *samplePtr;
  
  float sampleStep,sampleNo;
  int startSample,endSample,visibleSamples,windowSamples;
  int spectrum,bin,offset;
  short *centreSamplePtr;
  float *sgramPtr;
  float xmax,omega,dBcorrection;
  static float *fftBuffer=NULL,*windowData=NULL;
  
  wpi->fftSize=FFTSIZE;
  wpi->windowDuration=WINDOWLENGTH;
  
  /* various calcs */
  startSample=(int)(Gstart*(float)GsampleRate);
  endSample=(int)(Gend*(float)GsampleRate);
  
  if ((visibleSamples=endSample-startSample)<=0)
    return(0.);
  
  sampleStep=(float)visibleSamples/(float)NOSPECTRA;
  windowSamples=(int)(wpi->windowDuration*(float)GsampleRate);
  
  /* hamming window */
  myAllocate((void *)&windowData,windowSamples,FLOAT);
  omega=8.0*atan(1.0)/(float)windowSamples;
  for (sample=0,windowPtr=windowData;
       sample<windowSamples;
       sample++,windowPtr++)
    *windowPtr=(0.54-(0.46*cos((float)sample*omega)));
  
  /* fft/sgram space */
  myAllocate((void *)&fftBuffer,wpi->fftSize+2,FLOAT);
  /* allocate a tad more than is needed */
  if (wpi->sgram==NULL)
    myAllocate((void *)&wpi->sgram,(wpi->fftSize/2)*(NOSPECTRA+10),FLOAT);
  
  /* calculate correction factor */
  for (spectrum=0,sampleNo=(float)startSample,xmax=1.0;
       spectrum<NOSPECTRA;
       spectrum++,sampleNo+=sampleStep)
    {
      centreSamplePtr=&wpi->waveS[(int)(sampleNo+.5)];
      for (sample=0,samplePtr=centreSamplePtr-(windowSamples/2),xsum=0.0;
	   sample<windowSamples;
	   sample++,samplePtr++)
	xsum += *samplePtr * *samplePtr;
      
      if (xsum>xmax) xmax=xsum;
    }
  dBcorrection=((10.0*log10(xmax))+HEADROOM) - SPECRANGE;
  
  /* now perform fft */
  for (spectrum=0,sampleNo=(float)startSample,sgramPtr=wpi->sgram;
       spectrum<NOSPECTRA;
       spectrum++,sampleNo+=sampleStep)
    {
      centreSamplePtr=&wpi->waveS[(int)(sampleNo+.5)];

      offset=(wpi->fftSize-windowSamples)/2;
      if (offset<0) offset=0;
      
      /* pad before with zeros */
      for (sample=0,fftPtr=fftBuffer;
	   sample<offset;
	   sample++,fftPtr++)
	*fftPtr=0.;
      
      /* copy in speech */
      for (sample=0,windowPtr=windowData,
	   samplePtr=centreSamplePtr-(windowSamples/2);
	   sample<windowSamples;
	   sample++,fftPtr++,windowPtr++,samplePtr++)
	*fftPtr=*samplePtr * *windowPtr;
      
      /* pad rest of buffer */
      for (;
	   sample<=wpi->fftSize+2;
	   sample++,fftPtr++)
	*fftPtr=0.0;
      
      /* fft time */
      forwardFFT(fftBuffer,wpi->fftSize);
      /* greyscale time */
      for (bin=0,fftPtr=fftBuffer;
	   bin<wpi->fftSize/2;
	   bin++,sgramPtr++,fftPtr++)
	*sgramPtr=*fftPtr;
    }
  wpi->recalcSgram=FALSE;
  return (dBcorrection);
}

/*-------------------------------------------------------------------------*/

void calculateGreyscales (struct waveInfo *wpi, float dBcorrection)
     
{
  register int spectrum,bin,noBins;
  register float *sgramPtr,dBmagNorm;
  register char *greyScalePtr;
  register double *doublePtr;
  
  noBins=wpi->fftSize/2;
  
  if (wpi->greyScales==NULL)
    myAllocate((void *)&wpi->greyScales,(wpi->fftSize/2)*(NOSPECTRA+10),CHAR);
  
  if (wpi->sgramX==NULL)
    {
      myAllocate((void *)&wpi->sgramX,NOSPECTRA,DOUBLE);
      for (spectrum=0,doublePtr=wpi->sgramX;
	   spectrum<NOSPECTRA;
	   spectrum++,doublePtr++)
	*doublePtr=(double)spectrum/(double)NOSPECTRA;
    }
  
  if (wpi->sgramY==NULL)
    {
      myAllocate((void *)&wpi->sgramY,noBins,DOUBLE);
      for (bin=0,doublePtr=wpi->sgramY;
	   bin<noBins;
	   bin++,doublePtr++)
	*doublePtr=(double)bin/(double)noBins;
    }
  
  for (spectrum=0,sgramPtr=wpi->sgram,greyScalePtr=wpi->greyScales;
       spectrum<NOSPECTRA;
       spectrum++)
    {
      /* greyscale time */
      for (bin=0;
	   bin<noBins;
	   bin++,sgramPtr++,greyScalePtr++)
	{
	  if ((dBmagNorm=(*sgramPtr-dBcorrection)/SPECRANGE)<0.)
	    *greyScalePtr=(char)colourIndex[0];
	  else
	    *greyScalePtr=(char)
	      colourIndex[(int)((dBmagNorm*(float)(GREYSCALES-1))+.5)];
	}
    }
  wpi->recalcGreyScales=FALSE;
}

/*-------------------------------------------------------------------------*/

void waveformPaintProc (SUIT_object waveformDisplay)
     
{
  double lposn,rposn;
  struct waveInfo *wpi;
  
  wpi=&wfi[SUIT_getInteger(waveformDisplay,NUMBER)];
  if (!wpi->used)
    return;
  if (wpi->waveS==NULL)
    return;
  
  GP_setLineStyle(SOLID);
  GP_setLineWidth(1);
  GP_setColor(wpi->colour);
  GP_polyLineCoord(GdisplaySamples,GwaveX,wpi->waveD);
  
  /* draw the region start and end points */
  GP_setColor(GP_defColor("yellow",TRUE));
  GP_setWriteMode(WRITE_XOR);
  
  lposn=(Glcursor-Gstart)/(Gend-Gstart);
  GP_lineCoord(lposn,-1.0,lposn,1.0);
  rposn=(Grcursor-Gstart)/(Gend-Gstart);
  GP_lineCoord(rposn,-1.0,rposn,1.0);
  
  /*    if (wpi->noSamples>0)
	{
	sprintf(text,"%.3lfs",(double)wpi->start/(double)wpi->sampleRate);
	SUIT_setText(wpi->startTime,LABEL,text);
	sprintf(text,"%.3lfs",(double)wpi->end/(double)wpi->sampleRate);
	SUIT_setText(wpi->endTime,LABEL,text);
	}
	*/
  GP_setWriteMode(WRITE_REPLACE);
}

/*-------------------------------------------------------------------------*/

void zoomCallback (SUIT_object button)
     
{
  int frame,startSample,endSample,noSamples,sampleNo,sample;
  int startFrame,endFrame;

  /* moderate precise positions to align with frames */
  startFrame=(int)((Glcursor/GframeDuration)+.5);
  endFrame=(int)((Grcursor/GframeDuration)+.5);
  /* don't change if within one frame */
  if (startFrame==endFrame)
    return;

  GstartFrame=startFrame;
  GendFrame=endFrame;
  Gstart=GframeDuration*(float)GstartFrame;
  Glcursor=Gstart;
  Gend=GframeDuration*(float)GendFrame;
  Grcursor=Gend;
  
  /* calculate some new parameter display Xcoords */
  GdisplayFrames=GendFrame-GstartFrame;
  myAllocate((void *)&GparamX,GdisplayFrames,DOUBLE);
  for (frame=0;frame<GdisplayFrames;frame++)
    GparamX[frame]=(double)frame/(double)(GdisplayFrames-1);
  
  /* calculate some new waveform display Xcoords */
  startSample=(int)(Gstart*(float)GsampleRate);
  endSample=(int)(Gend*(float)GsampleRate);
  noSamples=(endSample-startSample);
  
  GsampleStep=1;
  while ( (int)((float)noSamples/(float)GsampleStep) > MAXPOINTS)
    GsampleStep++;
  GdisplaySamples=noSamples/GsampleStep;
  
  myAllocate((void *)&GwaveX,GdisplaySamples,DOUBLE);
  myAllocate((void *)&wfi[SYNTH].waveD,GdisplaySamples,DOUBLE);
  
  for (sampleNo=0,sample=startSample;
       sampleNo<GdisplaySamples;
       sampleNo++,sample+=GsampleStep)
    {
      GwaveX[sampleNo]=(double)sampleNo/(double)(GdisplaySamples-1);
      wfi[SYNTH].waveD[sampleNo]=(double)wfi[SYNTH].waveS[sample]/MAXSAMPLE;
    }
  wfi[SYNTH].recalcSgram=TRUE;
  wfi[SYNTH].recalcGreyScales=TRUE;
  wfi[SYNTH].recalcSpec=TRUE;
  SUIT_redisplayRequired(wfi[SYNTH].display);
  
  if (wfi[ORIG].used)
    {
      myAllocate((void *)&wfi[ORIG].waveD,GdisplaySamples,DOUBLE);
      for (sampleNo=0,sample=startSample;
	   sampleNo<GdisplaySamples;
	   sampleNo++,sample+=GsampleStep)
	wfi[ORIG].waveD[sampleNo]=(double)wfi[ORIG].waveS[sample]/MAXSAMPLE;
      wfi[ORIG].recalcSgram=TRUE;
      wfi[ORIG].recalcGreyScales=TRUE;
      wfi[ORIG].recalcSpec=TRUE;
      SUIT_redisplayRequired(wfi[ORIG].display);
    }
  
  SUIT_redisplayRequired(paramDisplay);
}

/*-------------------------------------------------------------------------*/

void unZoomCallback (SUIT_object button)
     
{
  int frame,startSample,endSample,noSamples;
  int sampleNo,sample;
  
  /* only unzoom if necessary */
/*  if ((Gstart==0.) && (Gend==Gduration))
    return;
*/  
  /* moderate precise positions to align with frames */
  GstartFrame=0;
  GendFrame=(int)(Gduration/GframeDuration);
  
  Gstart=Glcursor=0.;
  Gend=Grcursor=Gduration;
  
  /* calculate some new parameter display Xcoords */
  GdisplayFrames=GendFrame-GstartFrame;
  myAllocate((void *)&GparamX,GdisplayFrames,DOUBLE);
  for (frame=0;frame<GdisplayFrames;frame++)
    GparamX[frame]=(double)frame/(double)(GdisplayFrames-1);
  
  /* calculate some new waveform display Xcoords */
  startSample=(int)(Gstart*(float)GsampleRate);
  endSample=(int)(Gend*(float)GsampleRate);
  noSamples=(endSample-startSample);
  
  GsampleStep=1;
  while ( (int)((float)noSamples/(float)GsampleStep) > MAXPOINTS)
    GsampleStep++;
  GdisplaySamples=noSamples/GsampleStep;
  
  myAllocate((void *)&GwaveX,GdisplaySamples,DOUBLE);
  myAllocate((void *)&wfi[SYNTH].waveD,GdisplaySamples,DOUBLE);
  
  for (sampleNo=0,sample=startSample;
       sampleNo<GdisplaySamples;
       sampleNo++,sample+=GsampleStep)
    {
      GwaveX[sampleNo]=(double)sampleNo/(double)(GdisplaySamples-1);
      wfi[SYNTH].waveD[sampleNo]=(double)wfi[SYNTH].waveS[sample]/MAXSAMPLE;
    }
  wfi[SYNTH].recalcSgram=TRUE;
  wfi[SYNTH].recalcGreyScales=TRUE;
  wfi[SYNTH].recalcSpec=TRUE;
  SUIT_redisplayRequired(wfi[SYNTH].display);
  
  if (wfi[ORIG].used)
    {
      myAllocate((void *)&wfi[ORIG].waveD,GdisplaySamples,DOUBLE);
      for (sampleNo=0,sample=startSample;
	   sampleNo<GdisplaySamples;
	   sampleNo++,sample+=GsampleStep)
	wfi[ORIG].waveD[sampleNo]=(double)wfi[ORIG].waveS[sample]/MAXSAMPLE;
      wfi[ORIG].recalcSgram=TRUE;
      wfi[ORIG].recalcGreyScales=TRUE;
      wfi[ORIG].recalcSpec=TRUE;
      SUIT_redisplayRequired(wfi[ORIG].display);
    }
  
  SUIT_redisplayRequired(paramDisplay);
}

/*-------------------------------------------------------------------------*/

void rescaleCallback (SUIT_object button)
     
{
  SUIT_window windowS,windowO;

  windowS=SUIT_getWindow(wfi[SYNTH].display,WINDOW);
  windowO=SUIT_getWindow(wfi[ORIG].display,WINDOW);
  switch(SUIT_getInteger(button,CALLBACKARGUMENT))
    {
    case -1:
      windowS.bottom_left.y *= 2.0;
      windowS.top_right.y *= 2.0;
      windowO.bottom_left.y *= 2.0;
      windowO.top_right.y *= 2.0;
      break;

    case 1:
      windowS.bottom_left.y *= 0.5;
      windowS.top_right.y *= 0.5;
      windowO.bottom_left.y *= 0.5;
      windowO.top_right.y *= 0.5;
      break;

    case 0:
      windowS.bottom_left.y = -1.0;
      windowS.top_right.y = 1.0;
      windowO.bottom_left.y = -1.0;
      windowO.top_right.y = 1.0;
      break;
    }
  
  SUIT_setWindow(wfi[SYNTH].display,WINDOW,windowS);
  SUIT_setWindow(wfi[ORIG].display,WINDOW,windowO);
}

/*-------------------------------------------------------------------------*/

void waveformHitProc (SUIT_object waveformDisplay, SUIT_event myEvent)
     
{
  int actualSampleRate,startSample,endSample,playSamples;
  struct waveInfo *wpi;
  double gain,posn,absPosn;
  
  wpi=&wfi[SUIT_getInteger(waveformDisplay,NUMBER)];
  
  posn=myEvent.worldLocation.x;
  absPosn=Gstart+(posn*(Gend-Gstart));
  GP_setLineWidth(1);
  GP_setLineStyle(SOLID);

  switch(myEvent.type)
    {
    case CLICK:
      switch (myEvent.button)
	{
	case LEFT_BUTTON:
	  if (absPosn<Grcursor)
	    {
	      Glcursor=(float)absPosn;
	      wpi->recalcSpec=TRUE;
	      SUIT_redisplayRequired(wfi[SYNTH].display);
	      SUIT_redisplayRequired(wfi[ORIG].display);
	    }
	  break;
	  
	case MIDDLE_BUTTON:
	  goto play;
	  break;
	  
	case RIGHT_BUTTON:
	  if (absPosn>Glcursor)
	    {
	      Grcursor=(float)absPosn;
	      wpi->recalcSpec=TRUE;
	      SUIT_redisplayRequired(wfi[SYNTH].display);
	      SUIT_redisplayRequired(wfi[ORIG].display);
	    }
	  break;
	}
      break;
      
    case KEYSTROKE:
      switch (myEvent.keyboard)
	{
	case 's':
	case 'S':
	  if (wpi->recalcSpec)
	    calculateDFT(0);
	  break;
	  
	case 'c':
	case 'C':
	  if (wpi->recalcSpec)
	    calculateDFT(1);
	  break;
	  
	case ' ':
	  
	play:
	  /* ensure that audio device is open */
	  openAudioDevice();
	  startSample=(int)(Glcursor*(float)GsampleRate);
	  endSample=(int)(Grcursor*(float)GsampleRate);
	  playSamples=endSample-startSample;
	  /* read gain value */
	  gain=SUIT_getDouble(wpi->gainControl,CURRENT_VALUE);
	  
	  if (playSamples>0)
	    actualSampleRate=
	      outputAudioData(&wpi->waveS[startSample],playSamples,
			      GsampleRate,gain,outputPort);
	  
	  /* Warn if the sample rate was not that expected */
	  if (actualSampleRate!=GsampleRate)
	    {
	      sprintf(workSpace,"Sample rate:  Requested %d  Actual %d\n",
		      GsampleRate,actualSampleRate);
	      SUIT_inform(workSpace);
	    }
	  break;
	}
      
      
    case MOUSE_DOWN:
    case MOUSE_UP:
    case MOUSE_MOTION:
      /* do nothing */
      break;
    }
}

/*-------------------------------------------------------------------------*/

static float *fxContour=NULL;

void fxCallback (SUIT_object object)
     
{
  int noFrames,frame;
  float *fxPtr;
  double *dataPtr,range;
  
  if (wfi[ORIG].used)
    {
      noFrames=(int)(Gduration/GframeDuration);
      myAllocate((void *)&fxContour,noFrames+2,FLOAT);
      
      fxcep(wfi[ORIG].waveS,
	    GsampleRate,
	    noFrames,
	    GframeDuration,
	    fxContour);
      
      for (frame=0,dataPtr=variedParamInfo[FX].data,fxPtr=fxContour,
	   range=(double)variedParamInfo[FX].range;
	   frame<noFrames;
	   frame++,dataPtr++,fxPtr++)
	*dataPtr=(double)*fxPtr/range;
    }
  
  SUIT_setInteger(paramDisplay,DISPLAYMODE,PARAMETERS);
  SUIT_redisplayRequired(paramDisplay);
}

/*-------------------------------------------------------------------------*/

void spectrumCallback (SUIT_object object)
     
{
  int callbackArgument;
  
  callbackArgument=SUIT_getInteger(object,CALLBACKARGUMENT);
  calculateDFT(callbackArgument);
}

/*-------------------------------------------------------------------------*/


#define CEPCUT 0.002    /* cepstrum cut-off frequency = 500 Hz */

void calculateDFT (int smoothed)
     
{
  struct waveInfo *wpi;
  static float *fftbuf=NULL,*cepstbuf=NULL,*windowData=NULL;
  short *samplePtr;
  int i,j,pow,offset,bin,noSamples,fftSize,sample;
  int startSample,endSample,windowSample,windowSamples;
  float	omega,val,*xp,*rp;
  float xmax,xsum,dBcorrection,*windowPtr,*fftPtr;
  double log10(),cos(),atan(),log();
  
  /* how big is the region ? */
  startSample=(int)(Glcursor*(float)GsampleRate);
  endSample=(int)(Grcursor*(float)GsampleRate);
  if ((noSamples=(endSample-startSample))<=0)
    return;
  
  /* find power of two required */
  pow=8;fftSize=256;
  while (((noSamples) >> pow) > 0)
    {pow++;fftSize *= 2;}
  
  /* Allocate some workspace */
  myAllocate((void *)&fftbuf,fftSize+2,FLOAT);
  myAllocate((void *)&cepstbuf,fftSize+2,FLOAT);
  
  /* calculate dB correction factor */
  windowSamples=(int)(WINDOWLENGTH*(float)GsampleRate);
  for (wpi=&wfi[ORIG],xmax=0.0001;
       wpi<=&wfi[SYNTH];
       wpi++)
    {
      if (wpi->used)
	for (sample=0,samplePtr=&wpi->waveS[startSample],xmax=0.001;
	     sample<noSamples;
	     )
	  {
	    for (windowSample=0,xsum=0.0001;
		 windowSample<windowSamples;
		 windowSample++,sample++,samplePtr++)
	      xsum+= (float) (*samplePtr * *samplePtr);
	    if (xsum>xmax) xmax=xsum;
	  }
    }
  dBcorrection=((10.0*log10(xmax))+HEADROOM) - SPECRANGE;
  
  /* calculate hamming window */
  myAllocate((void *)&windowData,noSamples,FLOAT);
  omega=8.0*atan(1.0)/(float)noSamples;
  for (sample=0,windowPtr=windowData;
       sample<noSamples;
       sample++,windowPtr++)
    *windowPtr=(0.54-(0.46*cos((float)sample*omega)));
  
  for (wpi=&wfi[ORIG];wpi<=&wfi[SYNTH];wpi++)
    {
      if (wpi->used)
	{
	  wpi->specSize=fftSize/2;
	  myAllocate((void *)&wpi->spectrum,wpi->specSize,DOUBLE);
	  
	  offset=(fftSize-noSamples)/2;
	  if (offset<0) offset=0;
	  
	  /* pad before with zeros */
	  for (sample=0,fftPtr=fftbuf;
	       sample<offset;
	       sample++,fftPtr++)
	    *fftPtr=0.;
	  
	  /* copy in speech */
	  for (sample=0,windowPtr=windowData,
	       samplePtr=&wpi->waveS[startSample];
	       sample<noSamples;
	       sample++,fftPtr++,windowPtr++,samplePtr++)
	    *fftPtr=*samplePtr * *windowPtr;
	  
	  /* pad rest of buffer */
	  for (;
	       sample<fftSize;
	       sample++,fftPtr++)
	    *fftPtr=0.0;
	  
	  /* perform FFT */
	  ffta(fftbuf,fftSize);
	  
	  /* perform cepstral analysis if required */
	  if (smoothed)
	    {
	      rp = cepstbuf;
	      xp = fftbuf + fftSize + 1;
	      for (i=0;i<fftSize;i+=2)
		{		/* backwards spectrum */
		  val = 0.0001;
		  val += *xp * *xp;
		  xp--;
		  val += *xp * *xp;
		  xp--;
		  *rp++ = 10.0*log10(val);
		}
	      xp = cepstbuf + fftSize/2 - 1;
	      for (i=0;i<fftSize;i+=2)
		*rp++ = *xp--; /* forwards spectrum */
	      
	      /* perform real-to-complex fft */
	      ffta(cepstbuf,fftSize);
	      
	      /* zero out detail < 500Hz */
	      j = CEPCUT/(1.0/(float)GsampleRate);
	      xp = &cepstbuf[2*j+2];
	      for (i=j;i<fftSize;i++)
		{
		  *xp++ = 0.0;
		  *xp++ = 0.0;
		}
	      
	      /* do inverse fft */
	      ffts(cepstbuf,fftSize);
	      
	      /* copy smoothed energies into coefficient structure */
	      xp = cepstbuf + fftSize/2;
	      for (i=0;i<fftSize/2;i++)
		{
		  wpi->spectrum[i] = *xp++/SPECRANGE;
		  printf("bin %d  cep %.2f\n",bin,wpi->spectrum[i]);
		}

	    } /* if (smoothed) */
	  else
	    {
	      /* copy raw energies into coefficient structure */
	      xp = fftbuf+2;
	      for (i=0;i<fftSize/2;i++)
		{
		  val = 0.0001;
		  val += *xp * *xp;
		  xp++;
		  val += *xp * *xp;
		  xp++;
		  wpi->spectrum[i]=((10.0*log10(val))-dBcorrection)/SPECRANGE;
		  wpi->spectrum[i]/=2.;
		}
	    }
	}
      /* allocate some memory */
      myAllocate((void *)&wpi->specX,wpi->specSize,DOUBLE);
      
      for (bin=0;bin<wpi->specSize;bin++)
	wpi->specX[bin]=(double)bin/(double)wpi->specSize;
    }
  
  wpi->recalcSpec=FALSE;
  
  /* ensure that the window is correct */
  SUIT_setInteger(paramDisplay,DISPLAYMODE,SPECTRA);
  SUIT_setWindow(paramDisplay,WINDOW,
		 SUIT_getWindow(paramDisplay,DEFAULTWINDOW));
  SUIT_redisplayRequired(paramDisplay);
}

/*-------------------------------------------------------------------------*/

void synthesiseCallback (SUIT_object object)
     
{
  int param,frame,samplesPerFrame,sample,noFrames;
  int startSample,endSample,noSamples;
  short *dataShortPtr,*waveformDestination;
  double *dataDoublePtr;
  struct waveInfo *wpi;
  struct fpInfo *fpi;
  long *lptr;
  
  /* calculate samples per frame */
  samplesPerFrame=(int)(((double)fixedParamInfo[UI].currentValue/1000.)*
			(double)fixedParamInfo[SR].currentValue);
  
#ifdef KLATT88
  /* copy the data to the varieds accessed by parwav */
  
  /* first the constant parameters */
  for (param=0;param<NOFIXEDPARAMS;param++)
    spkrdef[param]=fixedParamInfo[param].currentValue;
  
  /* fill in the appropriate spkrdef value */
  spkrdef[1]=samplesPerFrame;
#endif
  
#ifdef KLATT80
  /* fill in the correct values for the global parameters */
  for (fpi=fixedParamInfo;
       fpi->name!=NULL;
       fpi++)
    {
      /* fill in current value in the frame */
      if (fpi->kframePtr!=NULL)
	*(fpi->kframePtr)=fpi->currentValue;
    }
  
  /* calculate the number of samples per frame */
  kglobals.nspfr=(long)((float)fixedParamInfo[SR].currentValue*
			((float)fixedParamInfo[UI].currentValue/1000.0));
  parwave_init(&kglobals);
#endif
  
  /* calculate the number of samples*/
  noFrames=variedParamInfo[0].noFrames;
  
  /* locate the waveform destination */
  wpi=&wfi[SYNTH];
  waveformDestination=wpi->waveS;
  
  /* mark that the waveform has changed */
  wpi->recalcSpec=TRUE;
  wpi->recalcSgram=TRUE;
  wpi->recalcGreyScales=TRUE;
  
  for (frame=0;
       frame<noFrames;
       frame++,waveformDestination+=samplesPerFrame)
    {
#ifdef KLATT88
      /* insert values into a Klatt Frame */
      for (param=0;param<NOVARIEDPARAMS;param++)
	pars[param]=(int)(variedParamInfo[param].data[frame]*
			  (double)variedParamInfo[param].range);
      
      /* move the next value of FX into the last pars entry */
      pars[NEXTFX]=(int)((double)variedParamInfo[FX].data[frame+1]*
			 (double)variedParamInfo[FX].range);
      
      /* Change to Klatt's 1/10Hz steps */
      pars[FX] *= 10;
      pars[NEXTFX] *= 10;
      
      parwav(waveformDestination);
#endif
#ifdef KLATT80
      /* insert values into a Klatt Frame */
      for (param=0,lptr=(long *)&kframe;
	   param<NOVARIEDPARAMS;
	   param++,lptr++)
	{
	  *lptr=(long)(variedParamInfo[param].data[frame]*
		       (double)variedParamInfo[param].range);
	}
      
      /* Change to Klatt's 1/10Hz steps */
      kframe.F0hz10 *=10;
      
      parwave(&kglobals,&kframe,waveformDestination);
#endif
    }
  
  /* calculate some new waveform display Xcoords */
  startSample=(int)(Gstart*(float)GsampleRate);
  endSample=(int)(Gend*(float)GsampleRate);
  noSamples=(endSample-startSample);
  
  GsampleStep=1;
  while ( ((int)(float)noSamples/(float)GsampleStep)>MAXPOINTS)
    GsampleStep++;
  GdisplaySamples=noSamples/GsampleStep;
  
  dataShortPtr=wpi->waveS+startSample;
  dataDoublePtr=wpi->waveD;
  for (sample=0;
       sample<GdisplaySamples;
       sample++,dataShortPtr+=GsampleStep,dataDoublePtr++)
    *dataDoublePtr=(double)*dataShortPtr/MAXSAMPLE;
  
  /* ensure that the new waveform is displayed */
  SUIT_redisplayRequired(wfi[SYNTH].display);
  /*  SUIT_redisplayRequired(paramDisplay);*/
}

/*-------------------------------------------------------------------------*/

void new (SUIT_object object)
     
{
  int reply;
  
  if (hasBeenEdited)
    {
      sprintf(workSpace,
	      "Do you want to abandon changes to %s ?",
	      currentFilename);
      reply=SUIT_askYesNo(workSpace);
      if (reply!=REPLY_YES)
	return;
    }
  
  /* destroy original waveform data */
  if (wfi[ORIG].used)
    {
      wfi[ORIG].used=FALSE;
      if (wfi[ORIG].sgram!=NULL) free(wfi[ORIG].sgram);
      if (wfi[ORIG].greyScales!=NULL) free(wfi[ORIG].greyScales);
      if (wfi[ORIG].waveS!=NULL) free(wfi[ORIG].waveS);
      if (wfi[ORIG].waveD!=NULL) free(wfi[ORIG].waveD);
      wfi[ORIG].sgram=NULL;
      wfi[ORIG].greyScales=NULL;
      wfi[ORIG].waveS=NULL;
      wfi[ORIG].waveD=NULL;
    }
  
  SUIT_redisplayRequired(wfi[ORIG].display);
  
  /* set to reasonable values, destroy old ones, and synthesise */
  initialise((int)(fixedParamInfo[DU].Default/fixedParamInfo[UI].Default),
	     fixedParamInfo[UI].Default,
	     fixedParamInfo[SR].Default,
	     TRUE,  
	     TRUE,  /* synthesise */ 
	     TRUE); /* display */
  
  currentFilename=SUIT_copyString("unnamed");
  hasBeenEdited=FALSE;
}

/*-------------------------------------------------------------------------*/

#define REVERSESHORT(s)	s = ((s & 0xFF00) >> 8) | ((s & 0x00FF) << 8);
#define REVERSELONG(l)	l = ((l & 0xFF000000L) >> 24) | ((l & 0x00FF0000L) >> 8) | ((l & 0x0000FF00L) << 8) | ((l & 0x000000FFL) << 24);

struct riff_file_header {
  char		riffid[4];	/* "RIFF" */
  unsigned long	size;		/* length of RIFF block */
} riffhead;

struct riff_wave_header {
  char		waveid[4];	/* "WAVE" */
  char		fmtid[4];	/* "fmt " */
  unsigned long	size;		/* 16 */
  unsigned short	format;		/* 1 */
  unsigned short	nchan;		/* 1 or 2 */
  unsigned long	srate;		/* srate */
  unsigned long	brate;		/* srate * #bytes/samp * nchan */
  unsigned short	balign;		/* 1 or 2 or 4 */
  unsigned short	nbits;		/* 8 or 16 */
} riffwavehead;

struct riff_data_header {
  char		dataid[4];	/* "data" */
  unsigned long	size;		/* length of signal in bytes */
} riffdatahead;

/*-------------------------------------------------------------------------*/

void writeRIFFheader(FILE *op, int nsample, int nchan, int do8bit)

{
  strcpy(riffhead.riffid,"RIFF");
  if (do8bit)
    riffhead.size = sizeof(struct riff_wave_header) +
      sizeof(struct riff_data_header) +
	nsample * nchan;
  else
    riffhead.size = sizeof(struct riff_wave_header) +
      sizeof(struct riff_data_header) +
	nsample * 2 * nchan;
  REVERSELONG(riffhead.size);
  fwrite((char *)&riffhead,sizeof(struct riff_file_header),1,op);

}

void writeRIFFwaveheader(FILE *op, int nsample, double srate, 
			 int nchan, int do8bit)
{
  strcpy(riffwavehead.waveid,"WAVE");
  strcpy(riffwavehead.fmtid,"fmt ");
  riffwavehead.size = 16;
  riffwavehead.format = 1;
  riffwavehead.nchan = nchan;
  riffwavehead.srate = (long)srate;
  riffwavehead.brate = (long)srate * ((do8bit)?1:2) * nchan;
  riffwavehead.balign = ((do8bit) ? 1 : 2)*nchan;
  riffwavehead.nbits = (do8bit) ? 8 : 16;
  
  REVERSELONG(riffwavehead.size);
  REVERSESHORT(riffwavehead.format);
  REVERSESHORT(riffwavehead.nchan);
  REVERSELONG(riffwavehead.srate);
  REVERSELONG(riffwavehead.brate);
  REVERSESHORT(riffwavehead.balign);
  REVERSESHORT(riffwavehead.nbits);
  
  fwrite((char *)&riffwavehead,sizeof(struct riff_wave_header),1,op);
}


void RIFFdataheader(FILE *op, int nsample, int nchan, int do8bit)
{
  strcpy(riffdatahead.dataid,"data");
  if (do8bit)
    riffdatahead.size = nsample * nchan;
  else
    riffdatahead.size = nsample * 2 * nchan;
  REVERSELONG(riffdatahead.size);
  fwrite((char *)&riffdatahead,sizeof(struct riff_data_header),1,op);
}

/*-------------------------------------------------------------------------*/

void loadAudioFile (SUIT_object button)
     
{
#ifdef SFS
  /* SFS Variables */
  struct item_header item;
#endif
  
  /* others */
  int sample,noFrames,fid,fileType;
  int sampleRate,newSampleRate,noSamples;
  char *filename=NULL;
  short *shortPtr=NULL;
  struct waveInfo *wpi;
  SUIT_object dbox;

  FILE *inFp=NULL;
  char wavMagic[4];
  long wavLong,len;
  int testInt=1,lsbFirst;
  char *lsbPtr;

  lsbPtr=(char *)&testInt;
  if (*lsbPtr==1)
    lsbFirst=TRUE;
  else 
    lsbFirst=FALSE;
  
#ifdef LINUX
  dbox = SUIT_createFileBrowserDialogBox("browser","./",
					 "Load an audio file",
					 "Open file named:");
#else /* SUN */
  dbox = SUIT_createFileBrowserDialogBox("browser",NULL,
					 "Load an audio file",
					 "Open file named:");
#endif
  fileType=SUIT_getInteger(button,CALLBACKARGUMENT);

  switch (fileType)
    {
    case WAVFILE:
      SUIT_setText(dbox,FILE_FILTER,"*.wav");
      break;
    case AUFILE:
      SUIT_setText(dbox,FILE_FILTER,"*.au");
      break;
#ifdef SFS
    case SFSFILE:
      SUIT_setText(dbox,FILE_FILTER,"*.sfs");
      break;
#endif
    }

  if (SUIT_activateDialogBox (dbox) == REPLY_OK)
    {
      filename=SUIT_copyString(SUIT_getText(dbox,CURRENT_VALUE));
      SUIT_destroyObject(dbox);
    }
  else
    {
      SUIT_destroyObject(dbox);
      return;
    }
  
  /* information about this item */
  wpi=&wfi[ORIG];
  
  switch (fileType)
    {
    case WAVFILE:
      if ((inFp=fopen(filename,"r"))==NULL)
	{
	  sprintf(workSpace,"Couldn't open %s",filename);
	  SUIT_inform(workSpace);
	  return;
	}

      /* read headers */
      fread((char *)&riffhead,sizeof(struct riff_file_header),1,inFp);
      if (strncmp(riffhead.riffid,"RIFF",4))
	{
	  sprintf(workSpace,"Couldn't read RIFF header in %s",filename);
	  SUIT_inform(workSpace);
	  return;
	}
      fread((char *)&riffwavehead,sizeof(struct riff_wave_header),1,inFp);
      if (strncmp(riffwavehead.waveid,"WAVE",4))
	{
	  sprintf(workSpace,"Couldn't read RIFF wave header in %s",filename);
	  SUIT_inform(workSpace);
	  return;
	}
      fread((char *)&riffdatahead,sizeof(struct riff_data_header),1,inFp);
      if (strncmp(riffdatahead.dataid,"data",4))
	{
	  sprintf(workSpace,"Couldn't read RIFF wave header in %s",filename);
	  SUIT_inform(workSpace);
	  return;
	}

      if (!lsbFirst)
	{
	  REVERSELONG(riffwavehead.srate);
	  REVERSELONG(riffdatahead.size);
	  REVERSESHORT(riffwavehead.nchan);
	  REVERSESHORT(riffwavehead.nbits);
	}

      noSamples=riffdatahead.size/(riffwavehead.nchan*(riffwavehead.nbits/8));
      sampleRate=riffwavehead.srate;

      myAllocate((void *)&wpi->waveS,noSamples+2,SHORT);
      for (sample=0,shortPtr=wpi->waveS;
	   sample<noSamples;
	   sample++,shortPtr++)
	{
	  if (riffwavehead.nbits==8)
	    {
	      *shortPtr=(getc(inFp)-0x80)<<8;
	      if (riffwavehead.nchan==2)
		getc(inFp);
	    }
	  else
	    {
	      *shortPtr=getc(inFp)+(getc(inFp)<<8);
	      if (riffwavehead.nchan==2)
		{getc(inFp);getc(inFp);}
	    }
	}
      break;
      
    case AUFILE:
      SUIT_setText(dbox,FILE_FILTER,"*.au");
      return;
      break;
      
#ifdef SFS
    case SFSFILE:
      /* Try to open the SFS file */
      if ((fid=sfsopen(filename,"r",NULL))<0)
	{
	  sprintf(workSpace,"Couldn't open %s",filename);
	  SUIT_inform(workSpace);
	  return;
	}
      
      /* Now try and find an item in that file */
      if (!sfsitem(fid,SP_TYPE,"0",&item))
	{
	  sprintf(workSpace,"Couldn't find a speech item in %s",filename);
	  SUIT_inform(workSpace);
	  return;
	}
      
      /* load the item in */
      sampleRate=(int)(1.0/item.frameduration);
      noSamples=item.numframes;
      /* allocate space */
      myAllocate((void *)&wpi->waveS,noSamples+10,SHORT);
      
      /* read in the data */
      sfsread(fid,0,noSamples,wpi->waveS);
      break;
#endif
    }
  
#ifdef SUNOS413
  /* resample if needed */
  if ((newSampleRate=nearestSunSampleRate(sampleRate))!=sampleRate)
    {
      sprintf(workSpace,
	      "Resampling to nearest Sun playback rate (from %d to %d Hz)",
	      sampleRate,newSampleRate);
      SUIT_inform(workSpace);

      noSamples=resampleWave(sampleRate,newSampleRate,
			     &wpi->waveS,noSamples);
      sampleRate=newSampleRate;
    }
#endif
  
  /* show the filename in the window */
  SUIT_setText(wpi->fileName,LABEL,filename);
  
  /* set global parameters to these new values */
  GsampleRate=sampleRate;
  Gduration=(float)noSamples/(float)GsampleRate;
  GframeDuration=(float)fixedParamInfo[UI].currentValue/1000.;
  noFrames=(int)(Gduration/GframeDuration);
  Gduration=(float)noFrames*GframeDuration;
  
  wpi->used=TRUE;
  wpi->recalcSpec=TRUE;
  wpi->recalcGreyScales=TRUE;
  wpi->recalcSgram=TRUE;
  fixedParamInfo[SR].currentValue=GsampleRate;
  
  /* ensure that synthetic version has the same length */
  initialise (noFrames,
	      fixedParamInfo[UI].currentValue,
	      GsampleRate,
	      FALSE,
	      TRUE,
	      TRUE);
  
  SUIT_redisplayRequired(wpi[ORIG].display);
}

/*-------------------------------------------------------------------------*/

void saveSfsFile (SUIT_object button, char *passedFilename)

{
  /* SFS Variables */
  struct item_header item;

  /* others */
  int noSamples,fid;
  long tim;
  char *filename,*directory,*nlptr;
  struct waveInfo *wpi;
  SUIT_object dbox;

  if (button!=NULL)
    {

#ifdef LINUX
  directory="./";
#else /* SUN */
  directory=NULL;
#endif

      dbox = SUIT_createFileBrowserDialogBox("browser",directory,
					     "Save data to sfs file",
					     "Save in file named:");
      SUIT_setText(dbox,FILE_FILTER,"*.sfs");
      
      if (SUIT_activateDialogBox (dbox) == REPLY_OK)
	{
	  filename=SUIT_getText(dbox,CURRENT_VALUE);
	  SUIT_destroyObject(dbox);
	}
      else
	{
	  SUIT_destroyObject(dbox);
	  return;
	}
    }
  else
    filename=passedFilename;

  /* information about this item */
  wpi=&wfi[SYNTH];

  /* check SFS file */
  if ((fid=sfsopen(filename,"w",NULL))< 0)
    {
      if ((fid=sfsopen(filename,"c",NULL))<0)
	{
	  if (button!=NULL)
	    {
	      sprintf(workSpace,"Couldn't open %s",filename);
	      SUIT_inform(workSpace);
	    }
	  else
	    fprintf(stderr,"Couldn't open %s\n",filename);
	  return;
	}
    }

  /* create an sfs header */
  sfsheader(&item,
	    SP_TYPE,
	    0,
	    2,
	    1,
	    (1.0/(double)GsampleRate),
	    0.0,
	    1,
	    0,
	    0);
  tim=time((long *)0);
  sprintf(item.history,"%s (%s, SR:%d, %s)",
	  PROGRAM,
	  currentFilename,
	  GsampleRate,
	  ctime(&tim));
  
  /* remove the NL from the history item */
  nlptr=item.history-1;
  while (*++nlptr!=0)
    if (*nlptr==NL)
      *nlptr=' ';

  /* open the output file */
  if ((fid=sfschannel(filename,&item))<0)
    {
      if (button!=NULL)
	{
	  sprintf(workSpace,"Couldn't open %s",filename);
	  SUIT_inform(workSpace);
	}
      else
	fprintf(stderr,"Couldn't open %s\n",filename);
      return;
    }

  /* write and update file */
  noSamples=(int)(Gduration * (float)GsampleRate);
  sfswrite(fid,noSamples,(char *)wpi->waveS);
  sfsupdate(filename);
}

/*-------------------------------------------------------------------------*/

int resampleWave (int sr, int newSr, short **wave, int noSamples)

{
  short *shortPtr;
  int newNoSamples,sample;
  float *waveFloat=NULL,*newWaveFloat=NULL,*floatPtr=NULL;

  if (newSr==sr)
    return(0);

  GP_setCursor(WATCH_CURSOR);

  /* first convert to floats */
  myAllocate((void *)&waveFloat,noSamples,FLOAT);
  for (sample=0,shortPtr=*wave,floatPtr=waveFloat;
       sample<noSamples;
       sample++,shortPtr++,floatPtr++)
    *floatPtr=0.95*(float)*shortPtr;
      
  /* resample */
  intdecE(waveFloat,noSamples,sr,&newWaveFloat,&newNoSamples,newSr);
      
  /* convert floats to shorts */
  myAllocate((void *)wave,newNoSamples,SHORT);
  for (sample=0,shortPtr=*wave,floatPtr=newWaveFloat;
       sample<newNoSamples;
       sample++,shortPtr++,floatPtr++)
    *shortPtr=(short)*floatPtr;
      
  /* free memory */
  free(waveFloat);
  free(newWaveFloat);
       
  GP_setCursor(STANDARD_CURSOR);

  return(newNoSamples);
}

/*-------------------------------------------------------------------------*/

void loadParFile (SUIT_object button, char *passedFilename)
     
{
  int noFrames,frame,param,reply;
  char *filename,*token,*paramPtr,*valuePtr,*following;
  struct fpInfo *fpi;
  FILE *inputfp;
  SUIT_object dbox;
  
  if (hasBeenEdited)
    {
      sprintf(workSpace,"Do you want to abandon changes to %s ?",currentFilename);
      reply=SUIT_askYesNo(workSpace);
      if (reply!=REPLY_YES)
	return;
    }
  
  if (button!=NULL)
    {
#ifdef LINUX
      dbox = SUIT_createFileBrowserDialogBox("browser","./",
					     "Load a parameter file",
					     "Open file named:");
#else /* SUN */
      dbox = SUIT_createFileBrowserDialogBox("browser",NULL,
					     "Load a parameter file",
					     "Open file named:");
#endif

#ifdef KLATT88
      SUIT_setText(dbox,FILE_FILTER,"*.par");
#endif
#ifdef KLATT80
      SUIT_setText(dbox,FILE_FILTER,"*.p80");
#endif
      
      if (SUIT_activateDialogBox (dbox) == REPLY_OK)
	{
	  filename=SUIT_copyString(SUIT_getText(dbox,CURRENT_VALUE));
	  SUIT_destroyObject(dbox);
	}
      else
	{
	  SUIT_destroyObject(dbox);
	  return;
	}
    } /* button !=NULL */
  else
    filename=passedFilename;
  
  if ((inputfp=fopen(filename,"r"))==NULL)
    {
      sprintf(workSpace,"Couldn't open %s",filename);
      if (button!=NULL)
	SUIT_inform(workSpace);
      else
	fprintf(stderr,"%s\n",workSpace);
      return;
    }
  
  /* show the filename in the window */
  if (button!=NULL)
    SUIT_setText(wfi[SYNTH].fileName,LABEL,filename);
  
  /* first count the number of frames in the file */
  noFrames=0;
  while (fgets(workSpace,400,inputfp)!=NULL)
    {
      /* ignore any comments */
      if (strchr(workSpace,'*')!=NULL) continue;
      noFrames++;
    }
  
  /* record the number of frames */
  for (param=0;param<NOVARIEDPARAMS;param++)
    variedParamInfo[param].noFrames=noFrames;
  
  /* re-read the file noting any changes to fixed parameter values */
  
  rewind(inputfp);
  frame=0;
  while (fgets(workSpace,400,inputfp)!=NULL)
    {
      /* either a comment or a fixed parameter value */
      if ((following=strchr(workSpace,'*'))!=NULL)
	{
	  /* forget it if it's a short line */
	  if (strlen(workSpace)<4) continue;
	  
	  /* otherwise parse it a little */
	  paramPtr=strtok(following+1," \t:=\n");
	  valuePtr=strtok(NULL," \t:=*\n");
	  
	  /* alter the fixed parameter value */
	  for (fpi=fixedParamInfo;
	       fpi->name!=NULL;
	       fpi++)
	    {
	      if (strcmp(paramPtr,fpi->name)==0)
		{
		  fpi->currentValue=atoi(valuePtr);
		  break;
		}
	    }
	}
      else
	break;
    }
  
  /* allocate space for this number of frames */
  if (button==NULL)
    initialise(noFrames,
	       fixedParamInfo[UI].currentValue,
	       fixedParamInfo[SR].currentValue,
	       FALSE,
	       FALSE,
	       FALSE);  /* NOTHING TO DISPLAY ON */
  else
    initialise(noFrames,
	       fixedParamInfo[UI].currentValue,
	       fixedParamInfo[SR].currentValue,
	       FALSE,
	       FALSE,
	       TRUE);  /* update display */
  
  /* re-read the file having allocated sufficient space for it using
     initialise, but this time store the values */
  rewind(inputfp);
  frame=0;
  while (fgets(workSpace,400,inputfp)!=NULL)
    {
      /* either a comment or a fixed parameter value */
      if (strchr(workSpace,'*')!=NULL)
	continue;
      
      /* parse the line, ignore the first token (time value) */
      strtok(workSpace," \t,");
      for (param=0;param<NOVARIEDPARAMS;param++)
	{
	  /* read the next token */
	  token=strtok(NULL," \t,");
	  /* convert it to a double and normalise it */
	  variedParamInfo[param].data[frame]=
	    (atof(token)/(double)variedParamInfo[param].range);
	  /* allow this value to be recalled if necessary */
	  variedParamInfo[param].memorisedData[frame]=
	    (atof(token)/(double)variedParamInfo[param].range);
	  variedParamInfo[param].noFrames=noFrames;
	}
      frame++;
    }
  fclose(inputfp);
  
  /* set up the new one */
  currentFilename=SUIT_copyString(filename);
  hasBeenEdited=FALSE;
  
  /* double synthesise to remove initial rubbish */
  synthesiseCallback(NULL);
  synthesiseCallback(NULL);
}

/*-------------------------------------------------------------------------*/

void saveParFile (SUIT_object button)
     
{
  int frame,reply;
  char *filename=NULL;
  FILE *outputfp;
  struct fpInfo *fpi;
  struct vpInfo *vpi;
  SUIT_object dbox;
  
  
  /* Ask for a filename */
  do
    {
#ifdef LINUX
      dbox = SUIT_createFileBrowserDialogBox("browser","./",
					     "Save a parameter file",
					     "Save a file named:");
#else /* SUN */
      dbox = SUIT_createFileBrowserDialogBox("browser",NULL,
					     "Save a parameter file",
					     "Save a file named:");
#endif

#ifdef KLATT88
      SUIT_setText(dbox,FILE_FILTER,"*.par");
#endif
#ifdef KLATT80
      SUIT_setText(dbox,FILE_FILTER,"*.p80");
#endif
      
      /* Ask for a filename */
      if (SUIT_activateDialogBox (dbox) == REPLY_OK)
	{
	  filename=SUIT_copyString(SUIT_getText(dbox,CURRENT_VALUE));
	  SUIT_destroyObject(dbox);
	}
      else
	{
	  SUIT_destroyObject(dbox);
	  return;
	}
      
      if (access(filename,F_OK)==0)
        {
          sprintf(workSpace,"Ok to overwrite %s ?",filename);
          reply=SUIT_askYesNoCancel(workSpace);
          if (reply==REPLY_CANCEL)  return;
          else if (reply==REPLY_NO) continue;
          else break;
        }
      else break;
    }
  while (TRUE);
  
  if ((outputfp=fopen(filename,"w"))==NULL)
    {
      sprintf(workSpace,"Couldn't open %s",filename);
      SUIT_inform(workSpace);
      return;
    }
  
  /* save results to a file */
  
  fprintf(outputfp,"/* Produced by %s */\n",PROGRAM);
  for (fpi=fixedParamInfo;
       fpi->name!=NULL;
       fpi++)
    fprintf(outputfp,"/*  %s : %ld */\n",fpi->name,fpi->currentValue);
  
  for (frame=0;
       frame<variedParamInfo[0].noFrames;
       frame++)
    {
      fprintf(outputfp,"%5ld: ",frame*fixedParamInfo[UI].currentValue);
      for (vpi=variedParamInfo;
	   vpi->name!=NULL;
	   vpi++)
	fprintf(outputfp,"%d ",(int)(vpi->data[frame]*(double)vpi->range));
      fprintf(outputfp,"\n");
    }
  fclose(outputfp);
  
  if (filename!=NULL)
    free(filename);
  
  /* modify the edited flag */
  hasBeenEdited=FALSE;
}

/*-------------------------------------------------------------------------*/

void defaultValueCallback (SUIT_object button)
     
{
  /* make a copy of the current synthesis values */
  int count,number;
  struct vpInfo *vpi;
  double *ptr,defaultValue;
  SUIT_object display;
  
  display=SUIT_getObject(button,DISPLAY);
  if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
    {
      vpi=&variedParamInfo[number];
      defaultValue=(double)vpi->Default/(double)vpi->range;
      for (count=0,ptr=vpi->data;
	   count<vpi->noFrames;
	   count++,ptr++)
	*ptr=defaultValue;
      
    }
  /* ensure that the display is updated */
  SUIT_redisplayRequired(display);
}

/*-------------------------------------------------------------------------*/

void memoriseParamCallback (SUIT_object button)
     
{
  /* make a copy of the current synthesis values */
  int count,number;
  struct vpInfo *vpi;
  double *from,*to;
  SUIT_object display;
  
  display=SUIT_getObject(button,DISPLAY);
  if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
    {
      vpi=&variedParamInfo[number];
      for (count=0,from=vpi->data,to=vpi->memorisedData;
	   count<vpi->noFrames;
	   count++,to++,from++)
	*to=*from;
    }
  /* ensure that the display is updated */
  SUIT_redisplayRequired(display);
}

/*-------------------------------------------------------------------------*/

void recallParamCallback (SUIT_object button)
     
{
  /* make a copy of the current synthesis values */
  int count,number;
  struct vpInfo *vpi;
  double *from,*to;
  SUIT_object display;
  
  display=SUIT_getObject(button,DISPLAY);
  if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
    {
      vpi=&variedParamInfo[number];
      for (count=0,from=vpi->memorisedData,to=vpi->data;
	   count<vpi->noFrames;
	   count++,to++,from++)
	*to=*from;
    }
  
  /* ensure that the display is updated */
  SUIT_redisplayRequired(display);
  
  /* reflect this change */
  synthesiseCallback(NULL);
}

/*-------------------------------------------------------------------------*/

void memoriseAllCallback (SUIT_object button)
     
{
  /* make a copy of the current synthesis values */
  int count;
  struct vpInfo *vpi;
  double *from,*to;
  
  for (vpi=variedParamInfo;
       vpi->name!=NULL;
       vpi++)
    {
      for (count=0,from=vpi->data,to=vpi->memorisedData;
	   count<vpi->noFrames;
	   count++,to++,from++)
	*to=*from;
    }
}

/*-------------------------------------------------------------------------*/

void recallAllCallback (SUIT_object button)
     
{
  /* make a copy of the current synthesis values */
  int count;
  struct vpInfo *vpi;
  double *from,*to;
  
  for (vpi=variedParamInfo;
       vpi->name!=NULL;
       vpi++)
    {
      for (count=0,from=vpi->memorisedData,to=vpi->data;
	   count<vpi->noFrames;
	   count++,to++,from++)
	*to=*from;
    }
  
  /* ensure that the display is updated */
  SUIT_redisplayRequired(SUIT_getObject(button,DISPLAY));
  
  /* reflect this change */
  synthesiseCallback(NULL);
}

/*-------------------------------------------------------------------------*/

void hideAllCallback (SUIT_object button)
     
{
  struct vpInfo *vpi;
  
  for (vpi=variedParamInfo;
       vpi->name!=NULL;
       vpi++)
    {
      vpi->state=NOTDISPLAYED;
      SUIT_setColor(vpi->button,BACKGROUND_COLOR,GP_defColor("gray",TRUE));
      SUIT_setColor(vpi->button,FOREGROUND_COLOR,GP_defColor("black",TRUE));
    }
  SUIT_setInteger(paramDisplay,EDITEDPARAM,-1);
  SUIT_redisplayRequired(paramDisplay);
}

/*-------------------------------------------------------------------------*/

void portCallback (SUIT_object object)
     
{
  if (SUIT_getBoolean(object,CURRENT_VALUE)==TRUE)
    outputPort |= SUIT_getInteger(object,CALLBACKARGUMENT);
  else
    outputPort &= ~SUIT_getInteger(object,CALLBACKARGUMENT);
}


/*-------------------------------------------------------------------------*/

void alterValueCallback (SUIT_object object)
     
{
  int frame,callbackArgument,number;
  double *dataPtr,adjustment;
  struct vpInfo *vpi;
  SUIT_object display;
  
  /* find out which parameter is being edited */
  display=SUIT_getObject(object,DISPLAY);
  if ((number=SUIT_getInteger(display,EDITEDPARAM))>=0)
    {
      vpi=&variedParamInfo[number];
      
      /* get the extra callback parameter */
      callbackArgument=SUIT_getInteger(object,CALLBACKARGUMENT);
      
      /* take note of the change */
      hasBeenEdited=TRUE;
      
      /* now adjust the parameter */
      adjustment=(double)callbackArgument * 0.01;
      for (frame=0,dataPtr=vpi->data;
	   frame<vpi->noFrames;
	   frame++,dataPtr++)
	{
	  /* adjust value ensuring that it does not go out of range */
	  *dataPtr += adjustment;
	  if (*dataPtr <0.0) *dataPtr=0.0;
	  if (*dataPtr >1.0) *dataPtr=1.0;
	}
      
      
      /* update display */
      SUIT_redisplayRequired(paramDisplay);
    }
}

/*-------------------------------------------------------------------------*/

void exitProgram (SUIT_object button)
     
{
  int reply;
  
  if (hasBeenEdited)
    {
      sprintf(workSpace,
	      "Do you want to abandon changes to %s ?",currentFilename);
      reply=SUIT_askYesNo(workSpace);
      if (reply!=REPLY_YES)
	return;
      SUIT_done(DO_NOT_SAVE_SUI_FILE,EXIT_APPLICATION);
    }
  else
    if (SUIT_askYesNo("Are you sure you want to exit?")==REPLY_YES)
      {
	SUIT_done(DO_NOT_SAVE_SUI_FILE,EXIT_APPLICATION);
      }
}

/*-------------------------------------------------------------------------*/

/* MAIN */

void main (int argc, char *argv[])
     
{
  unsigned short rgb;
  int arg,greyScale;
  char *parameterFile=NULL,*sfsFile=NULL;
  GP_color colour;
  
  /* intercept any command line arguments */
  for (arg=0; arg<argc; arg++)
    {
      if (argv[arg][0]=='-')
	{
	  switch (argv[arg][1])
	    {
	    case 'p':
	      /* load and process file */
	      parameterFile=argv[arg+1];
	      if (access(parameterFile,R_OK)<0)
		{
		  fprintf(stderr,
			  "\nCan't read the parameter file: %s\n\n",
			  parameterFile);
		  exit(1);
		}
	      break;
	      
	    case 's':
	      /* load and process file */
	      sfsFile=argv[arg+1];
	      break;
	      
	    case 'h':
	      fprintf(stderr,"\n%s -p parameter file  -s sfs file\n\n",
		      PROGRAM);
	      break;
	    }
	}
    }
  
  /* if given files as arguments, then process and save */
  if (sfsFile && parameterFile)
    {
      loadParFile(NULL,parameterFile);
      synthesiseCallback(NULL);
      saveSfsFile(NULL,sfsFile);
      exit(0);
    }
  
  /* carry on and initialise application */
  
  /* totally arbitrary */
  SRGP_setMaxPointlistSize(MAXPOINTS);
  
#ifdef NOSUIFILE    
  SUIT_initFromCode(argv[0]);
#else
  SUIT_init(argv[0]);
#endif
  
  createObjects();
  initialise((int)(fixedParamInfo[DU].Default/
		   fixedParamInfo[UI].Default),
	     fixedParamInfo[UI].Default,
	     fixedParamInfo[SR].Default,
	     TRUE,
	     TRUE,  /* synthesise */
	     TRUE); /* display */
  
  /* load the colour table with the colours */
  for (greyScale=0;greyScale<GREYSCALES;greyScale++)
    {
      rgb=rgbValues[greyScale]*256;
      SUIT_setColor(paramDisplay,"dummyColour",
		    GP_defColorRGB(rgb,rgb,rgb,TRUE));
      colour=SUIT_getColor(paramDisplay,"dummyColour");
      colourIndex[greyScale]=GP_getColorIndex(colour.colorName);
    }
  SUIT_beginStandardApplication();
}

/*-------------------------------------------------------------------------*/
