/*-------------------------------------------------------------------------

VARIOUS SIGNAL PROCESSING FUNCTIONS

-------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>

void forwardFFT (float *data, int n);
void ffta (float *data, int n);
void ffts (float *data, int n);
void realft(float *data, int n, int isign);
void four1(float *data, int nn, int isign);
void fxcep (short *waveform, int sampleRate, int noFrames, 
	    float frameStep, float *fxContour);
void cepstrum(float *alpha, float *omega);
void c_rules(float *,int *, float *,float *,float,int);
int findpeak(float *curve,int lowlimit,int highlimit);

/*-------------------------------------------------------------------------*/

/* fftBuffer should have fftSize+2 floats' space allocated */

void forwardFFT (float *fftBuffer, int fftSize)

{
  register float mag, *fftPtr, *magSpecPtr;
  register int bin;

  realft(fftBuffer-1,fftSize/2,1);

  fftBuffer[fftSize]=fftBuffer[1];
  fftBuffer[fftSize+1]=0.0;
  fftBuffer[1]=0.0;

  /* calculate amplitude spectrum, put it in first fftSize/2 elements */
  for (bin=0,fftPtr=fftBuffer+2,magSpecPtr=fftBuffer;
       bin<fftSize/2;
       bin++)
    {
      mag=0.0001;
      mag+= (*fftPtr * *fftPtr);fftPtr++;
      mag+= (*fftPtr * *fftPtr);fftPtr++;
      *magSpecPtr++=10.*log10(mag);
    }
}

/*-------------------------------------------------------------------------*/

/************************************************ 
  Complex Discrete Inverse Fast Fourier Transform, 
  ************************************************/  


void ffts(float *x, int points)
{       
  long i, n;                   
  
  n = points >> 1;
  x[1] = x[points];              /* last frequency component put in x[1] */
  realft(x-1, n, -1);
  for (i = 0; i < n; i++) 
    x[i] /= n;
}  

/**************************************** 
  Complex Discrete Fast Fourier Transform, 
  ***************************************/  

void ffta(float *x, int points)
{ 
  long n;                   
  
  n = points >> 1;
  realft(x-1, n, 1);
  x[points]   = x[1];
  x[points+1] = (float)0.0;  
  x[1]        = (float)0.0;
}  

/*-------------------------------------------------------------------------*/

/* REALFT - Numerical Recipes in C */
/*
This routine takes a single real function and calculates either its
DFT or its Inverse DFT depending on 'isign' as for 'four1'. 
On Entry: 'data[1..2n]' contains the real function of length '2n',
          where 'n' MUST be an integral power of 2. 
On Exit:  'data[1..2n]' contains the positive half of the complex DFT.
          'data[1]' contains the real valued first component and
          'data[2]' contains the real valued last component of the
          complex transform. 
This routine also calculates the inverse transform of a complex data
array if it is the transform of real data and 'isign' is set to -1.
(Result in this case must be multiplied by 1/n.)
*/

void realft(float *data, int n, int isign)
     
{
  int i,i1,i2,i3,i4,n2p3;
  float c1=(float)0.5,c2,h1r,h1i,h2r,h2i;
  double wr,wi,wpr,wpi,wtemp,theta;
  
  /* Initialise the recurrence. */
  theta = 3.141592653589793/(double) n;
  if (isign == 1) {
    c2 = (float)-0.5; 
    four1(data, n, 1);                 /* Forward Transform */
  } 
  else {
    c2 = (float)0.5;
    theta = -theta;
  }           
  
  wtemp = sin(0.5*theta);
  wpr   = -2.0*wtemp*wtemp;
  wpi   = sin(theta);
  wr    = 1.0+wpr;
  wi    = wpi;
  n2p3  = 2*n+3;  
  
  for (i = 2; i <= n/2; i++) {        /* Case i==1 done below */
    
    /* Separate transforms from the data */
    i4  = 1+(i3=n2p3-(i2=1+(i1=i+i-1)));
    h1r = c1*(data[i1]+data[i3]);
    h1i = c1*(data[i2]-data[i4]);
    h2r = -c2*(data[i2]+data[i4]);
    h2i = c2*(data[i1]-data[i3]);
    
    /* Recombine to form true transform of original data */
    data[i1] = (float)(h1r+wr*h2r-wi*h2i);
    data[i2] = (float)(h1i+wr*h2i+wi*h2r);
    data[i3] = (float)(h1r-wr*h2r+wi*h2i);
    data[i4] = (float)(-h1i+wr*h2i+wi*h2r);
    
    /* The recurrence Relations */
    wr = (wtemp = wr)*wpr-wi*wpi+wr;
    wi = wi*wpr+wtemp*wpi+wi;
  }    
  
  if (isign == 1) { 
    /* Squeeze the first & last data into the original array */ 
    data[1] = (h1r=data[1])+data[2];
    data[2] = h1r-data[2];
  } 
  else {
    data[1] = c1*((h1r=data[1])+data[2]);
    data[2] = c1*(h1r-data[2]);
    four1(data, n, -1);                /* Inverse Transform */
  }
}

/* FOUR1 - Numerical Recipes in C */

/*
  FFT routine based on one originally written by N. Brenner of Lincoln
  Laboratories.  
  On entry: 'data[1..2*nn]' contains 'nn' complex data points, with each
  complex value occupying two consecutive locations, real
  value stored in the lower address.  'nn' MUST be an integer
  power of 2 (this is not checked for).   
  On exit:  'data[1..2*nn]' contains the complex transformation output,
  again stored in pairs of real and imaginary values.
  'isign' determins the direction of the transform: 
  1 for forward DFT,
  -1 for inverse DFT, scaled by a factor of 'nn'; 
  (i.e. not normalised by 1/N).
  */

#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr
  
  void four1(float *data, int nn, int isign)

{
  long n,mmax,m,j,istep,i;
  double wtemp,wr,wpr,wpi,wi,theta; /* double precision for the */
  /* trigonometric recurrences. */
  float tempr,tempi;
  
  n=nn << 1;
  j=1;
  
  /* Bit-Reversal routine */
  for (i = 1; i < n; i += 2) {
    if (j > i) {
      SWAP(data[j],data[i]);   /* exchange the two complex numbers */
      SWAP(data[j+1],data[i+1]);
    }
    m = n >> 1;
    while (m >= 2 && j > m) {
      j -= m;
      m >>= 1;
    }
    j += m;
  }                 
  
  /* Danielson-Lanczos loops */
  mmax = 2;
  while (n > mmax) {      
    /* outer loop executed ln(nn) times */
    istep = 2*mmax; 
    
    /* initialise for the trigonometric recurrence */
    theta = 6.28318530717959/(isign*mmax); 
    wtemp = sin(0.5*theta);
    wpr   = -2.0*wtemp*wtemp;
    wpi   = sin(theta);
    wr    = 1.0;
    wi    = 0.0;
    
    /* two nested inner loops */
    for (m = 1; m < mmax; m += 2) {
      for (i = m; i <= n; i += istep) {   
	
	/* Danielson-Lanczos formula: */
	j          = i+mmax;
	tempr      = (float)(wr*data[j]-wi*data[j+1]);
	tempi      = (float)(wr*data[j+1]+wi*data[j]);
	data[j]    = data[i]-tempr;
	data[j+1]  = data[i+1]-tempi;
	data[i]   += tempr;
	data[i+1] += tempi;
      }                                 
      
      /* trigonometric recurrence */
      wr = (wtemp=wr)*wpr-wi*wpi+wr;
      wi = wi*wpr+wtemp*wpi+wi;
    }
    mmax = istep;
  }
}

#undef SWAP

/*-------------------------------------------------------------------------*/

/* fxcep -- fundamental frequency estimation by cepstral analysis, after Noll */

/* cep version 1.0 - L C Whitaker  20 august 1985         */
/* cep version 2.0 - D.M. Howard  March 1986              */
/* fxcep version 3.0 - M.A.Huckvale March 1991
	- SFS version
	- Non array-processor version
	- tidied code
*/

/* manifest constants */
#define FRAME     	0.04		/* frame length */
#define HALFMAX 	256          	/* fft is symmetrical about HALFMAX */
#define MAXVEC  	512           	/* number of points in fft */
#define TWICEMAX 	1024		/* buffer size */
#define MAX_SAMP_RATE	20000.0		/* max sampling rate */
#define THRESHOLD	50.0		/* cepstral peak threshold */

#define MEMORY		3		/* frame memory for peak following */
#define THIS		2		/* this frame */
#define LAST		1		/* last frame */
#define PREVLAST	0		/* frame before last */

void fxcep (short *waveform, int sampleRate, int noFrames, 
	    float frameStep, float *fxContour)

{
  float	alpha[TWICEMAX];		/* fft buffer */
  float	omega[TWICEMAX];		/* cepstrum buffer */
  float	window[TWICEMAX];		/* hamming window */
  int	ptime[MEMORY];			/* peak tracking memory */
  float	pfreq[MEMORY];
  float	pamp[MEMORY];
  int frame,stepSamples,j;
  register int sample,frameSamples;
  register short *windowStartPtr,*samplePtr;
  register float *bufferPtr,*windowPtr,*fxPtr;
  float constant;

  /* hamming window coefficients */
  if ((frameSamples=(int)(FRAME*(float)sampleRate))>MAXVEC)
    frameSamples=MAXVEC;

  constant=8.0*atan(1.0)/(float)frameSamples;
  for (sample=0,windowPtr=window;
       sample<frameSamples;
       sample++,windowPtr++)
    *windowPtr=(0.54-(0.46*cos((float)sample*constant)));

  for (;
       sample< MAXVEC;
       sample++,windowPtr++) 
    *windowPtr=0.;

  /* now process speech in windows */
  stepSamples=(int)(frameStep*(float)sampleRate);

  for (frame=0,windowStartPtr=waveform,fxPtr=fxContour;
       frame<noFrames;
       frame++,windowStartPtr+=stepSamples,fxPtr++)
    {
      for (sample=0,samplePtr=windowStartPtr,bufferPtr=alpha,windowPtr=window;
	   sample<frameSamples;
	   sample++,samplePtr++,bufferPtr++,windowPtr++)
	*bufferPtr=(float)*samplePtr * *windowPtr;

      for (;
	   sample<MAXVEC;
	   sample++,samplePtr++,bufferPtr++,windowPtr++)
	*bufferPtr=(float)*samplePtr * *windowPtr;
	   
      /* calculate cepstrum */
      cepstrum(alpha,omega);
      
      /* apply Noll's rules to cepstrum */
      ptime[THIS]=0;
      pamp[THIS]=0;
      pfreq[THIS]=0;
      c_rules(omega,ptime,pamp,pfreq,THRESHOLD,sampleRate);

      /* output decision on last frame */
      *fxPtr= pfreq[LAST];

      /* shift memory */
      for (j=0;j<MEMORY-1;j++)
	{
	  ptime[j] = ptime[j+1];
	  pamp[j] = pamp[j+1];
	  pfreq[j] = pfreq[j+1];
	}
    }
    /* *fxPtr=pfreq[LAST]; */
}


/* cepstal analysis */
void cepstrum(float *alpha, float *omega)

{
  register int 	i;
  int	siz=MAXVEC;
  
  /* perform real fft */
  ffta(alpha,siz);
  
  /* find log magnitude spectrum */
  memset(omega,0,TWICEMAX*sizeof(float));
  omega[HALFMAX] = log(alpha[0]*alpha[0] + 1.0E-6);
  for (i=0;i<(HALFMAX-1);i++) {
    /* reflected log magnitude */
    omega[HALFMAX-i-1] = 
      omega[HALFMAX+i+1] = log(alpha[2*i+2]*alpha[2*i+2] +
			       alpha[2*i+3]*alpha[2*i+3] + 1.0E-6);
  }
  
  /* second fft */
  ffta(omega,siz);
  
  /* calculate magnitude */
  for (i=0;i<HALFMAX;i++)
    omega[i] = (omega[2*i]*omega[2*i] + omega[2*i+1]*omega[2*i+1])/MAXVEC;
  
}

/* ************* */
/* * c_rules.c * */
/* ************* */
/* *************************** */
/* * (c) SPAR GROUP May 1986 * */
/* *************************** */
/******************************************************/
/*   Automatic tracking of cepstral peaks             */
/*   Using Noll's algotithm                           */
/*   L C Whitaker    24 September 1985                */
/******************************************************/

void c_rules(omega,pt,amplitude,pfreq,threshold,s_rate)
     float   *omega;     /* cepstral data */
     int     *pt;        /* position of cepstral peak */
     float   *amplitude; /* amplitude of peak */
     float   *pfreq;     /* pitch estimate */
     float   threshold;      /* threshold for cepstral peaks */
     int	s_rate;		/* sampling rate */
{
  float    gradient;  /* gradient of linear weighting */
  float    intersect; /* linear weighting at time zero */
  float    line[MAXVEC]; /* linear weighting */
  float    templevel; /* temporary threshold */
  int      track=0;
  int      newpeak;
  int      lowlimit,highlimit;
  int      point;
  int      j,ms_5;
  float    temp1,temp2;
  float    b;
  
  /* initial values */
  gradient =  4.0*1000.0/14.0;
  intersect = 10.0/14.0;
  lowlimit = 2.0*s_rate/1000; /* 2ms in samples */
  highlimit = 15*s_rate/1000; /* 15 ms in samples */
  templevel = threshold;
  point = 0;
  
  for(j=0;j<HALFMAX;j++){
    line[j] = (gradient*j/(float)s_rate) + intersect;
  }
  
  /* linear weighting */
  for(j=0;j<HALFMAX;j++){
    omega[j] = omega[j] * line[j];
  }
  
  /* detect maximum peak in cepstrum */
  pt[THIS] = findpeak(omega,lowlimit,highlimit);
  j = pt[THIS];
  amplitude[THIS] = omega[j];
  if (track == 1) {
    b = (float)(pt[THIS] - pt[LAST]);
    if (b < 0.0) b = -b;
    if (b/(float)s_rate <= 1.0/1000.0) templevel = threshold/2.0;
  }
  
  /* check if amplitude of cepstral peak is above threshold */
  if (amplitude[THIS] >= templevel)
    point = 3;
  else 
    point = 0;
  
  /* check for pitch doubling */
  if ((track == 1) && (pt[THIS] > 0) && (pt[LAST] > 0)) {
    temp1 = (float)s_rate/(float)pt[THIS];
    temp2 = (float)s_rate/(float)pt[LAST];
    if (temp2 >= 1.6*temp1) {
      ms_5 = 5*s_rate/10000;
      newpeak = findpeak(omega,(pt[THIS]/2) -ms_5,(pt[THIS]/2) +ms_5);
      b = (float)(newpeak - pt[LAST]);
      if (b < 0.0) b = -b;
      if (b/(float)s_rate <= 1.0/(1000.0)) templevel = threshold/2.0;
      else templevel = threshold;
      if (omega[newpeak] >= templevel){
	pt[THIS] = newpeak;
	amplitude[THIS] = omega[newpeak];
      }
    }
  }
  
  /* does previous peak exceed threshold */
  if (amplitude[LAST] >= templevel) {
    if (point == 3)
      point += 2;
    else
      point += 1;
  }
  if ((point ==3) || (point ==1)) {
    /* is peak at PREVLAST cepstrum a pitch peak */
    if (pfreq[PREVLAST] > 20.0) point += 1;
  }
  
  /* use switch to determine pitch of LAST frame */
  switch(point){
  case 0 :
    pfreq[LAST] = 0.0;
    pt[LAST] = 0;
    /* unvoiced frame */
    break;
  case 1 :
    pfreq[LAST] = 0.0;
    pt[LAST] = 0;
    /* isolated pitch peak - unvoiced frame */
    break;
  case 2 :
    track = 0;
    pfreq[LAST] = (float)s_rate/pt[LAST];
    /* voiced segment - stop pitch tracking */
    break;
  case 3 :
    pfreq[LAST] = 0.0;
    pt[LAST] = 0;
    /* unvoiced frame */
    break;
  case 4 :
    pt[LAST] = (pt[PREVLAST]+pt[THIS])/2;
    pfreq[LAST] = (float)s_rate/pt[LAST];
    /* isolated absence of pitch peak */
    break;
  case 5 :
    pfreq[LAST] = (float)s_rate/pt[LAST];
    track = 1;
    /* voiced frame - start or continue tracking */
    break;
    default :
      fprintf(stderr,"AARGH! error in cepstrules()");
    break;
  }
}

int findpeak(curve,lowlimit,highlimit)
     float 	*curve;
     int 	lowlimit;
     int 	highlimit;
{
  int 	i;
  int	maxpos=0;
  float	maxamp=0.0;
  
  if (lowlimit > MAXVEC) return(0);
  if (highlimit > MAXVEC) highlimit = MAXVEC;
  
  for (i=lowlimit;i<highlimit;i++)
    if (curve[i] > maxamp) {
      maxamp = curve[i];
      maxpos = i;
    }
  
  return(maxpos);
}
